# -*- coding: utf-8 -*-

# python std lib
from __future__ import with_statement

import asyncio

import pytest

# 3rd party imports
from mock import patch

# rediscluster imports
from coredis import StrictRedisCluster
from coredis.exceptions import AskError
from coredis.pool import ClusterConnectionPool
from coredis.utils import b


class DummyConnectionPool(ClusterConnectionPool):
    pass


class DummyConnection:
    pass


@pytest.mark.asyncio
async def test_pipeline_ask_redirection():
    """
    Test that the server handles ASK response when used in pipeline.

    At first call it should return a ASK ResponseError that will point
    the client to the next server it should talk to.

    Important thing to verify is that it tries to talk to the second node.
    """
    r = StrictRedisCluster(host="127.0.0.1", port=7000)
    with patch.object(StrictRedisCluster, "parse_response") as parse_response:

        async def response(connection, *args, **options):
            async def response(connection, *args, **options):
                async def response(connection, *args, **options):
                    assert connection.host == "127.0.0.1"
                    assert connection.port == 7001

                    return "MOCK_OK"

                parse_response.side_effect = response
                raise AskError("12182 127.0.0.1:7001")

            parse_response.side_effect = response
            raise AskError("12182 127.0.0.1:7001")

        parse_response.side_effect = response

        p = await r.pipeline()
        await p.connection_pool.initialize()
        p.connection_pool.nodes.nodes["127.0.0.1:7001"] = {
            "host": u"127.0.0.1",
            "server_type": "master",
            "port": 7001,
            "name": "127.0.0.1:7001",
        }
        await p.set("foo", "bar")
        assert await p.execute() == ["MOCK_OK"]


@pytest.mark.asyncio
async def test_moved_redirection():
    """
    Test that the client handles MOVED response.

    At first call it should return a MOVED ResponseError that will point
    the client to the next server it should talk to.

    Important thing to verify is that it tries to talk to the second node.
    """
    r0 = StrictRedisCluster(host="127.0.0.1", port=7000)
    r2 = StrictRedisCluster(host="127.0.0.1", port=7002)
    await r0.flushdb()
    await r2.flushdb()
    assert await r0.set("foo", "bar")
    assert await r2.get("foo") == b"bar"


@pytest.mark.asyncio
async def test_moved_redirection_pipeline(monkeypatch):
    """
    Test that the server handles MOVED response when used in pipeline.

    At first call it should return a MOVED ResponseError that will point
    the client to the next server it should talk to.

    Important thing to verify is that it tries to talk to the second node.
    """
    r0 = StrictRedisCluster(host="127.0.0.1", port=7000)
    r2 = StrictRedisCluster(host="127.0.0.1", port=7002)
    await r0.flushdb()
    await r2.flushdb()
    p = await r0.pipeline()
    await p.set("foo", "bar")
    assert await p.execute() == [True]
    assert await r2.get("foo") == b"bar"


async def assert_moved_redirection_on_slave(sr, connection_pool_cls, cluster_obj):
    """ """
    # we assume this key is set on 127.0.0.1:7000(7003)
    await sr.set("foo16706", "foo")
    await asyncio.sleep(1)
    with patch.object(connection_pool_cls, "get_node_by_slot") as return_slave_mock:
        return_slave_mock.return_value = {
            "name": "127.0.0.1:7004",
            "host": "127.0.0.1",
            "port": 7004,
            "server_type": "slave",
        }

        master_value = {
            "host": "127.0.0.1",
            "name": "127.0.0.1:7000",
            "port": 7000,
            "server_type": "master",
        }
        with patch.object(
            ClusterConnectionPool, "get_master_node_by_slot"
        ) as return_master_mock:
            return_master_mock.return_value = master_value
            assert await cluster_obj.get("foo16706") == b("foo")
            assert return_slave_mock.call_count == 1


@pytest.mark.asyncio
async def test_moved_redirection_on_slave_with_default_client(sr):
    """
    Test that the client is redirected normally with default
    (readonly_mode=False) client even when we connect always to slave.
    """
    await assert_moved_redirection_on_slave(
        sr,
        ClusterConnectionPool,
        StrictRedisCluster(host="127.0.0.1", port=7000, reinitialize_steps=1),
    )


@pytest.mark.asyncio
async def test_moved_redirection_on_slave_with_readonly_mode_client(sr):
    """
    Ditto with READONLY mode.
    """
    await assert_moved_redirection_on_slave(
        sr,
        ClusterConnectionPool,
        StrictRedisCluster(
            host="127.0.0.1", port=7000, readonly=True, reinitialize_steps=1
        ),
    )


@pytest.mark.asyncio
async def test_access_correct_slave_with_readonly_mode_client(sr):
    """
    Test that the client can get value normally with readonly mode
    when we connect to correct slave.
    """

    # we assume this key is set on 127.0.0.1:7000(7003)
    await sr.set("foo16706", "foo")
    await asyncio.sleep(1)

    with patch.object(ClusterConnectionPool, "get_node_by_slot") as return_slave_mock:
        return_slave_mock.return_value = {
            "name": "127.0.0.1:7004",
            "host": "127.0.0.1",
            "port": 7004,
            "server_type": "slave",
        }

        master_value = {
            "host": "127.0.0.1",
            "name": "127.0.0.1:7000",
            "port": 7000,
            "server_type": "master",
        }
        with patch.object(
            ClusterConnectionPool, "get_master_node_by_slot", return_value=master_value
        ):
            readonly_client = StrictRedisCluster(
                host="127.0.0.1", port=7000, readonly=True
            )
            assert b("foo") == await readonly_client.get("foo16706")
            readonly_client = StrictRedisCluster.from_url(
                url="redis://127.0.0.1:7000/0", readonly=True
            )
            assert b("foo") == await readonly_client.get("foo16706")



# -*- coding: utf-8 -*-

# python std lib
from __future__ import with_statement

import asyncio
import time

# 3rd party imports
import pytest

# rediscluster imports
from coredis import StrictRedis, StrictRedisCluster
from coredis.exceptions import ConnectionError


async def wait_for_message(pubsub, timeout=0.5, ignore_subscribe_messages=False):
    now = time.time()
    timeout = now + timeout
    while now < timeout:
        message = await pubsub.get_message(
            ignore_subscribe_messages=ignore_subscribe_messages, timeout=0.01
        )
        if message is not None:
            return message
        await asyncio.sleep(0.01)
        now = time.time()
    return None


def make_message(type, channel, data, pattern=None):
    return {
        "type": type,
        "pattern": pattern and pattern.encode("utf-8") or None,
        "channel": channel.encode("utf-8"),
        "data": data.encode("utf-8") if isinstance(data, str) else data,
    }


def make_subscribe_test_data(pubsub, type):
    if type == "channel":
        return {
            "p": pubsub,
            "sub_type": "subscribe",
            "unsub_type": "unsubscribe",
            "sub_func": pubsub.subscribe,
            "unsub_func": pubsub.unsubscribe,
            "keys": ["foo", "bar", "uni" + chr(4456) + "code"],
        }
    elif type == "pattern":
        return {
            "p": pubsub,
            "sub_type": "psubscribe",
            "unsub_type": "punsubscribe",
            "sub_func": pubsub.psubscribe,
            "unsub_func": pubsub.punsubscribe,
            "keys": ["f*", "b*", "uni" + chr(4456) + "*"],
        }
    assert False, "invalid subscribe type: {0}".format(type)


class TestPubSubSubscribeUnsubscribe:
    async def _test_subscribe_unsubscribe(
        self, p, sub_type, unsub_type, sub_func, unsub_func, keys
    ):
        for key in keys:
            assert await sub_func(key) is None

        # should be a message for each channel/pattern we just subscribed to
        for i, key in enumerate(keys):
            assert await wait_for_message(p) == make_message(sub_type, key, i + 1)

        for key in keys:
            assert await unsub_func(key) is None

        # should be a message for each channel/pattern we just unsubscribed
        # from
        for i, key in enumerate(keys):
            i = len(keys) - 1 - i
            assert await wait_for_message(p) == make_message(unsub_type, key, i)

    @pytest.mark.asyncio
    async def test_channel_subscribe_unsubscribe(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "channel")
        await self._test_subscribe_unsubscribe(**kwargs)

    @pytest.mark.asyncio
    async def test_pattern_subscribe_unsubscribe(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "pattern")
        await self._test_subscribe_unsubscribe(**kwargs)

    async def _test_resubscribe_on_reconnection(
        self, p, sub_type, sub_func, keys, *args, **kwargs
    ):
        for key in keys:
            assert await sub_func(key) is None

        # should be a message for each channel/pattern we just subscribed to
        for i, key in enumerate(keys):
            assert await wait_for_message(p) == make_message(sub_type, key, i + 1)

        # manually disconnect
        p.connection.disconnect()

        # calling get_message again reconnects and resubscribes
        # note, we may not re-subscribe to channels in exactly the same order
        # so we have to do some extra checks to make sure we got them all
        messages = []
        for i, _ in enumerate(keys):
            messages.append(await wait_for_message(p))

        unique_channels = set()
        assert len(messages) == len(keys)
        for i, message in enumerate(messages):
            assert message["type"] == sub_type
            assert message["data"] == i + 1
            assert isinstance(message["channel"], bytes)
            channel = message["channel"].decode("utf-8")
            unique_channels.add(channel)

        assert len(unique_channels) == len(keys)
        for channel in unique_channels:
            assert channel in keys

    @pytest.mark.asyncio
    async def test_resubscribe_to_channels_on_reconnection(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "channel")
        await self._test_resubscribe_on_reconnection(**kwargs)

    @pytest.mark.asyncio
    async def test_resubscribe_to_patterns_on_reconnection(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "pattern")
        await self._test_resubscribe_on_reconnection(**kwargs)

    async def _test_subscribed_property(
        self, p, sub_type, unsub_type, sub_func, unsub_func, keys
    ):
        assert p.subscribed is False
        await sub_func(keys[0])
        # we're now subscribed even though we haven't processed the
        # reply from the server just yet
        assert p.subscribed is True
        assert await wait_for_message(p) == make_message(sub_type, keys[0], 1)
        # we're still subscribed
        assert p.subscribed is True

        # unsubscribe from all channels
        await unsub_func()
        # we're still technically subscribed until we process the
        # response messages from the server
        assert p.subscribed is True
        assert await wait_for_message(p) == make_message(unsub_type, keys[0], 0)
        # now we're no longer subscribed as no more messages can be delivered
        # to any channels we were listening to
        assert p.subscribed is False

        # subscribing again flips the flag back
        await sub_func(keys[0])
        assert p.subscribed is True
        assert await wait_for_message(p) == make_message(sub_type, keys[0], 1)

        # unsubscribe again
        await unsub_func()
        assert p.subscribed is True
        # subscribe to another channel before reading the unsubscribe response
        await sub_func(keys[1])
        assert p.subscribed is True
        # read the unsubscribe for key1
        assert await wait_for_message(p) == make_message(unsub_type, keys[0], 0)
        # we're still subscribed to key2, so subscribed should still be True
        assert p.subscribed is True
        # read the key2 subscribe message
        assert await wait_for_message(p) == make_message(sub_type, keys[1], 1)
        await unsub_func()
        # haven't read the message yet, so we're still subscribed
        assert p.subscribed is True
        assert await wait_for_message(p) == make_message(unsub_type, keys[1], 0)
        # now we're finally unsubscribed
        assert p.subscribed is False

    @pytest.mark.asyncio
    async def test_subscribe_property_with_channels(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "channel")
        await self._test_subscribed_property(**kwargs)

    @pytest.mark.asyncio
    async def test_subscribe_property_with_patterns(self, r):
        kwargs = make_subscribe_test_data(r.pubsub(), "pattern")
        await self._test_subscribed_property(**kwargs)

    @pytest.mark.asyncio
    async def test_ignore_all_subscribe_messages(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)

        checks = (
            (p.subscribe, "foo"),
            (p.unsubscribe, "foo"),
            # (p.psubscribe, 'f*'),
            # (p.punsubscribe, 'f*'),
        )

        assert p.subscribed is False
        for func, channel in checks:
            assert await func(channel) is None
            assert p.subscribed is True
            assert await wait_for_message(p) is None
        assert p.subscribed is False

    @pytest.mark.asyncio
    async def test_ignore_individual_subscribe_messages(self, r):
        p = r.pubsub()

        checks = (
            (p.subscribe, "foo"),
            (p.unsubscribe, "foo"),
            # (p.psubscribe, 'f*'),
            # (p.punsubscribe, 'f*'),
        )

        assert p.subscribed is False
        for func, channel in checks:
            assert await func(channel) is None
            assert p.subscribed is True
            message = await wait_for_message(p, ignore_subscribe_messages=True)
            assert message is None
        assert p.subscribed is False


class TestPubSubMessages:
    """
    Bug: Currently in cluster mode publish command will behave different then in
         standard/non cluster mode. See (docs/Pubsub.md) for details.

         Currently StrictRedis instances will be used to test pubsub because they
         are easier to work with.
    """

    def get_strict_redis_node(self, port, host="127.0.0.1"):
        return StrictRedis(port=port, host=host)

    def setup_method(self, *args):
        self.message = None

    def message_handler(self, message):
        self.message = message

    @pytest.mark.asyncio
    async def test_published_message_to_channel(self):
        node = self.get_strict_redis_node(7000)
        p = node.pubsub(ignore_subscribe_messages=True)
        await p.subscribe("foo")

        assert await node.publish("foo", "test message") == 1

        message = await wait_for_message(p)
        assert isinstance(message, dict)
        assert message == make_message("message", "foo", "test message")

        # Cleanup pubsub connections
        p.close()

    @pytest.mark.asyncio
    @pytest.mark.xfail(reason="This test is buggy and fails randomly")
    async def test_publish_message_to_channel_other_server(self):
        """
        Test that pubsub still works across the cluster on different nodes
        """
        node_subscriber = self.get_strict_redis_node(7000)
        p = node_subscriber.pubsub(ignore_subscribe_messages=True)
        await p.subscribe("foo")

        node_sender = self.get_strict_redis_node(7001)
        # This should return 0 because of no connected clients to this server.
        assert await node_sender.publish("foo", "test message") == 0

        message = await wait_for_message(p)
        assert isinstance(message, dict)
        assert message == make_message("message", "foo", "test message")

        # Cleanup pubsub connections
        p.close()

    @pytest.mark.asyncio
    @pytest.mark.xfail(reason="Pattern pubsub do not work currently")
    async def test_published_message_to_pattern(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        try:
            await p.subscribe("foo")
            await p.psubscribe("f*")
            # 1 to pattern, 1 to channel
            assert await r.publish("foo", "test message") == 2

            message1 = await wait_for_message(p)
            message2 = await wait_for_message(p)
            assert isinstance(message1, dict)
            assert isinstance(message2, dict)

            expected = [
                make_message("message", "foo", "test message"),
                make_message("pmessage", "foo", "test message", pattern="f*"),
            ]

            assert message1 in expected
            assert message2 in expected
            assert message1 != message2
        finally:
            await p.unsubscribe("foo")
            await p.punsubscribe("f*")

    @pytest.mark.asyncio
    async def test_channel_message_handler(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        try:
            await p.subscribe(foo=self.message_handler)
            assert await r.publish("foo", "test message") == 1
            assert await wait_for_message(p) is None
            assert self.message == make_message("message", "foo", "test message")
        finally:
            await p.unsubscribe("foo")

    @pytest.mark.asyncio
    @pytest.mark.xfail(reason="Pattern pubsub do not work currently")
    async def test_pattern_message_handler(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        await p.psubscribe(**{"f*": self.message_handler})
        assert await r.publish("foo", "test message") == 1
        assert await wait_for_message(p) is None
        assert self.message == make_message(
            "pmessage", "foo", "test message", pattern="f*"
        )

    @pytest.mark.asyncio
    @pytest.mark.xfail(reason="Pattern pubsub do not work currently")
    async def test_unicode_channel_message_handler(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        channel = "uni" + chr(4456) + "code"
        channels = {channel: self.message_handler}
        print(channels)
        await p.subscribe(**channels)
        assert await r.publish(channel, "test message") == 1
        assert await wait_for_message(p) is None
        assert self.message == make_message("message", channel, "test message")

    @pytest.mark.asyncio
    @pytest.mark.xfail(reason="Pattern pubsub do not work currently")
    async def test_unicode_pattern_message_handler(self, r):
        p = r.pubsub(ignore_subscribe_messages=True)
        pattern = "uni" + chr(4456) + "*"
        channel = "uni" + chr(4456) + "code"
        await p.psubscribe(**{pattern: self.message_handler})
        assert await r.publish(channel, "test message") == 1
        assert await wait_for_message(p) is None
        assert self.message == make_message(
            "pmessage", channel, "test message", pattern=pattern
        )


class TestPubSubAutoDecoding:
    "These tests only validate that we get unicode values back"

    channel = "uni" + chr(4456) + "code"
    pattern = "uni" + chr(4456) + "*"
    data = "abc" + chr(4458) + "123"

    def make_message(self, type, channel, data, pattern=None):
        return {"type": type, "channel": channel, "pattern": pattern, "data": data}

    def setup_method(self, *args):
        self.message = None

    def message_handler(self, message):
        self.message = message

    @pytest.mark.asyncio
    async def test_channel_subscribe_unsubscribe(self, o):
        p = o.pubsub()
        await p.subscribe(self.channel)
        assert await wait_for_message(p) == self.make_message(
            "subscribe", self.channel, 1
        )

        await p.unsubscribe(self.channel)
        assert await wait_for_message(p) == self.make_message(
            "unsubscribe", self.channel, 0
        )

    @pytest.mark.asyncio
    @pytest.mark.xfail(reason="Pattern pubsub do not work currently")
    async def test_pattern_subscribe_unsubscribe(self, o):
        p = o.pubsub()
        await p.psubscribe(self.pattern)
        assert await wait_for_message(p) == self.make_message(
            "psubscribe", self.pattern, 1
        )

        await p.punsubscribe(self.pattern)
        assert await wait_for_message(p) == self.make_message(
            "punsubscribe", self.pattern, 0
        )

    @pytest.mark.asyncio
    async def test_channel_publish(self, o):
        p = o.pubsub(ignore_subscribe_messages=True)
        await p.subscribe(self.channel)
        await o.publish(self.channel, self.data)
        assert await wait_for_message(p) == self.make_message(
            "message", self.channel, self.data
        )

    @pytest.mark.asyncio
    @pytest.mark.xfail(reason="Pattern pubsub do not work currently")
    async def test_pattern_publish(self, o):
        p = o.pubsub(ignore_subscribe_messages=True)
        await p.psubscribe(self.pattern)
        await o.publish(self.channel, self.data)
        assert await wait_for_message(p) == self.make_message(
            "pmessage", self.channel, self.data, pattern=self.pattern
        )

    @pytest.mark.asyncio
    async def test_channel_message_handler(self, o):
        p = o.pubsub(ignore_subscribe_messages=True)
        await p.subscribe(**{self.channel: self.message_handler})
        await o.publish(self.channel, self.data)
        assert await wait_for_message(p) is None
        assert self.message == self.make_message("message", self.channel, self.data)

        # test that we reconnected to the correct channel
        p.connection.disconnect()
        assert await wait_for_message(p) is None  # should reconnect
        new_data = self.data + "new data"
        await o.publish(self.channel, new_data)
        assert await wait_for_message(p) is None
        assert self.message == self.make_message("message", self.channel, new_data)

    @pytest.mark.asyncio
    @pytest.mark.xfail(reason="Pattern pubsub do not work currently")
    async def test_pattern_message_handler(self, o):
        p = o.pubsub(ignore_subscribe_messages=True)
        await p.psubscribe(**{self.pattern: self.message_handler})
        await o.publish(self.channel, self.data)
        assert await wait_for_message(p) is None
        assert self.message == self.make_message(
            "pmessage", self.channel, self.data, pattern=self.pattern
        )

        # test that we reconnected to the correct pattern
        p.connection.disconnect()
        assert await wait_for_message(p) is None  # should reconnect
        new_data = self.data + "new data"
        await o.publish(self.channel, new_data)
        assert await wait_for_message(p) is None
        assert self.message == self.make_message(
            "pmessage", self.channel, new_data, pattern=self.pattern
        )


class TestPubSubRedisDown:
    @pytest.mark.asyncio
    async def test_channel_subscribe(self, r):
        r = StrictRedis(host="localhost", port=6390)
        p = r.pubsub()
        with pytest.raises(ConnectionError):
            await p.subscribe("foo")


def test_pubsub_thread_publish():
    """
    This test will never fail but it will still show and be viable to use
    and to test the threading capability of the connectionpool and the publish
    mechanism.
    """
    startup_nodes = [{"host": "127.0.0.1", "port": "7000"}]

    r = StrictRedisCluster(
        startup_nodes=startup_nodes,
        max_connections=16,
        max_connections_per_node=16,
    )

    async def t_run(rc):
        for i in range(0, 50):
            await rc.publish("foo", "bar")
            await rc.publish("bar", "foo")
            await rc.publish("asd", "dsa")
            await rc.publish("dsa", "asd")
            await rc.publish("qwe", "bar")
            await rc.publish("ewq", "foo")
            await rc.publish("wer", "dsa")
            await rc.publish("rew", "asd")

        # Use this for debugging
        # print(rc.connection_pool._available_connections)
        # print(rc.connection_pool._in_use_connections)
        # print(rc.connection_pool._created_connections)

    try:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(asyncio.gather(*(t_run(r) for _ in range(10))))
    except Exception as e:
        print(e)
        print("Error: unable to start thread")


#
#
# class TestPubSubPubSubSubcommands:
#     """
#     Test Pub/Sub subcommands of PUBSUB
#     @see https://redis.io/commands/pubsub
#     """
#
#     async def test_pubsub_channels(self, r):
#         r.pubsub(ignore_subscribe_messages=True).subscribe('foo', 'bar', 'baz', 'quux')
#         channels = sorted(r.pubsub_channels())
#         assert channels == [b('bar'), b('baz'), b('foo'), b('quux')]
#
#     def test_pubsub_numsub(self, r):
#         r.pubsub(ignore_subscribe_messages=True).subscribe('foo', 'bar', 'baz')
#         r.pubsub(ignore_subscribe_messages=True).subscribe('bar', 'baz')
#         r.pubsub(ignore_subscribe_messages=True).subscribe('baz')
#
#         channels = [(b('bar'), 2), (b('baz'), 3), (b('foo'), 1)]
#         assert channels == sorted(r.pubsub_numsub('foo', 'bar', 'baz'))
#
#     def test_pubsub_numpat(self, r):
#         r.pubsub(ignore_subscribe_messages=True).psubscribe('*oo', '*ar', 'b*z')
#         assert r.pubsub_numpat() == 3



import asyncio
import inspect
import os
import socket
import ssl
import sys
import time
import warnings
from io import BytesIO

from coredis.exceptions import (
    AskError,
    BusyLoadingError,
    ClusterCrossSlotError,
    ClusterDownError,
    ConnectionError,
    ExecAbortError,
    InvalidResponse,
    MovedError,
    NoScriptError,
    ReadOnlyError,
    RedisError,
    ResponseError,
    TimeoutError,
    TryAgainError,
)
from coredis.utils import LOOP_DEPRECATED, b, nativestr

try:
    import hiredis

    HIREDIS_AVAILABLE = True
except ImportError:
    HIREDIS_AVAILABLE = False

SYM_STAR = b("*")
SYM_DOLLAR = b("$")
SYM_CRLF = b("\r\n")
SYM_LF = b("\n")
SYM_EMPTY = b("")


async def exec_with_timeout(coroutine, timeout, *, loop=None):
    try:
        if LOOP_DEPRECATED:
            return await asyncio.wait_for(coroutine, timeout)
        else:
            return await asyncio.wait_for(coroutine, timeout, loop=loop)
    except asyncio.TimeoutError as exc:
        raise TimeoutError(exc)


class SocketBuffer:
    def __init__(self, stream_reader, read_size):
        self._stream = stream_reader
        self.read_size = read_size
        self._buffer = BytesIO()
        # number of bytes written to the buffer from the socket
        self.bytes_written = 0
        # number of bytes read from the buffer
        self.bytes_read = 0

    @property
    def length(self):
        return self.bytes_written - self.bytes_read

    async def _read_from_socket(self, length=None):
        buf = self._buffer
        buf.seek(self.bytes_written)
        marker = 0

        try:
            while True:
                data = await self._stream.read(self.read_size)
                # an empty string indicates the server shutdown the socket
                if isinstance(data, bytes) and len(data) == 0:
                    raise ConnectionError("Socket closed on remote end")
                buf.write(data)
                data_length = len(data)
                self.bytes_written += data_length
                marker += data_length

                if length is not None and length > marker:
                    continue
                break
        except socket.error:
            e = sys.exc_info()[1]
            raise ConnectionError("Error while reading from socket: %s" % (e.args,))

    async def read(self, length):
        length = length + 2  # make sure to read the \r\n terminator
        # make sure we've read enough data from the socket
        if length > self.length:
            await self._read_from_socket(length - self.length)

        self._buffer.seek(self.bytes_read)
        data = self._buffer.read(length)
        self.bytes_read += len(data)

        # purge the buffer when we've consumed it all so it doesn't
        # grow forever
        if self.bytes_read == self.bytes_written:
            self.purge()

        return data[:-2]

    async def readline(self):
        buf = self._buffer
        buf.seek(self.bytes_read)
        data = buf.readline()
        while not data.endswith(SYM_CRLF):
            # there's more data in the socket that we need
            await self._read_from_socket()
            buf.seek(self.bytes_read)
            data = buf.readline()

        self.bytes_read += len(data)

        # purge the buffer when we've consumed it all so it doesn't
        # grow forever
        if self.bytes_read == self.bytes_written:
            self.purge()

        return data[:-2]

    def purge(self):
        self._buffer.seek(0)
        self._buffer.truncate()
        self.bytes_written = 0
        self.bytes_read = 0

    def close(self):
        try:
            self.purge()
            self._buffer.close()
        except Exception:
            # redis-py issue #633 suggests the purge/close somehow raised a
            # BadFileDescriptor error. Perhaps the client ran out of
            # memory or something else? It's probably OK to ignore
            # any error being raised from purge/close since we're
            # removing the reference to the instance below.
            pass
        self._buffer = None
        self._sock = None


class BaseParser:
    """Plain Python parsing class"""

    EXCEPTION_CLASSES = {
        "ERR": {"max number of clients reached": ConnectionError},
        "EXECABORT": ExecAbortError,
        "LOADING": BusyLoadingError,
        "NOSCRIPT": NoScriptError,
        "READONLY": ReadOnlyError,
        "ASK": AskError,
        "TRYAGAIN": TryAgainError,
        "MOVED": MovedError,
        "CLUSTERDOWN": ClusterDownError,
        "CROSSSLOT": ClusterCrossSlotError,
    }

    def parse_error(self, response):
        """Parse an error response"""
        error_code = response.split(" ")[0]
        if error_code in self.EXCEPTION_CLASSES:
            response = response[len(error_code) + 1 :]
            exception_class = self.EXCEPTION_CLASSES[error_code]
            if isinstance(exception_class, dict):
                exception_class = exception_class.get(response, ResponseError)
            return exception_class(response)
        return ResponseError(response)


class PythonParser(BaseParser):
    def __init__(self, read_size):
        self._stream = None
        self._buffer = None
        self._read_size = read_size
        self.encoding = None

    def __del__(self):
        try:
            self.on_disconnect()
        except Exception:
            pass

    def on_connect(self, connection):
        """Called when the stream connects"""
        self._stream = connection._reader
        self._buffer = SocketBuffer(self._stream, self._read_size)
        if connection.decode_responses:
            self.encoding = connection.encoding

    def on_disconnect(self):
        """Called when the stream disconnects"""
        if self._stream is not None:
            self._stream = None
        if self._buffer is not None:
            self._buffer.close()
            self._buffer = None
        self.encoding = None

    def can_read(self):
        return self._buffer and bool(self._buffer.length)

    async def read_response(self):
        if not self._buffer:
            raise ConnectionError("Socket closed on remote end")
        response = await self._buffer.readline()
        if not response:
            raise ConnectionError("Socket closed on remote end")

        byte, response = chr(response[0]), response[1:]

        if byte not in ("-", "+", ":", "$", "*"):
            raise InvalidResponse("Protocol Error: %s, %s" % (str(byte), str(response)))

        # server returned an error
        if byte == "-":
            response = response.decode()
            error = self.parse_error(response)
            # if the error is a ConnectionError, raise immediately so the user
            # is notified
            if isinstance(error, ConnectionError):
                raise error
            # otherwise, we're dealing with a ResponseError that might belong
            # inside a pipeline response. the connection's read_response()
            # and/or the pipeline's execute() will raise this error if
            # necessary, so just return the exception instance here.
            return error
        # single value
        elif byte == "+":
            pass
        # int value
        elif byte == ":":
            response = int(response)
        # bulk response
        elif byte == "$":
            length = int(response)
            if length == -1:
                return None
            response = await self._buffer.read(length)
        # multi-bulk response
        elif byte == "*":
            length = int(response)
            if length == -1:
                return None
            response = []
            for i in range(length):
                response.append(await self.read_response())
        if isinstance(response, bytes) and self.encoding:
            response = response.decode(self.encoding)
        return response


class HiredisParser(BaseParser):
    """Parser class for connections using Hiredis"""

    def __init__(self, read_size):
        if not HIREDIS_AVAILABLE:
            raise RedisError("Hiredis is not installed")
        self._stream = None
        self._reader = None
        self._read_size = read_size

    def __del__(self):
        try:
            self.on_disconnect()
        except Exception:
            pass

    def can_read(self):
        if not self._reader:
            raise ConnectionError("Socket closed on remote end")

        if self._next_response is False:
            self._next_response = self._reader.gets()
        return self._next_response is not False

    def on_connect(self, connection):
        self._stream = connection._reader
        kwargs = {
            "protocolError": InvalidResponse,
            "replyError": ResponseError,
        }
        if connection.decode_responses:
            kwargs["encoding"] = connection.encoding
        self._reader = hiredis.Reader(**kwargs)
        self._next_response = False

    def on_disconnect(self):
        if self._stream is not None:
            self._stream = None
        self._reader = None
        self._next_response = False

    async def read_response(self):
        if not self._stream:
            raise ConnectionError("Socket closed on remote end")

        # _next_response might be cached from a can_read() call
        if self._next_response is not False:
            response = self._next_response
            self._next_response = False
            return response

        response = self._reader.gets()
        while response is False:
            try:
                buffer = await self._stream.read(self._read_size)
            # CancelledError will be caught by client so that command won't be retried again
            # For more detailed discussion please see https://github.com/alisaifee/coredis/issues/56
            except asyncio.CancelledError:
                raise
            except Exception:
                e = sys.exc_info()[1]
                raise ConnectionError(
                    "Error {} while reading from stream: {}".format(type(e), e.args)
                )
            if not buffer:
                raise ConnectionError("Socket closed on remote end")
            self._reader.feed(buffer)
            response = self._reader.gets()
        if isinstance(response, ResponseError):
            response = self.parse_error(response.args[0])
        return response


if HIREDIS_AVAILABLE:
    DefaultParser = HiredisParser
else:
    DefaultParser = PythonParser


class RedisSSLContext:
    def __init__(self, keyfile=None, certfile=None, cert_reqs=None, ca_certs=None):
        self.keyfile = keyfile
        self.certfile = certfile
        if cert_reqs is None:
            self.cert_reqs = ssl.CERT_NONE
        elif isinstance(cert_reqs, str):
            CERT_REQS = {
                "none": ssl.CERT_NONE,
                "optional": ssl.CERT_OPTIONAL,
                "required": ssl.CERT_REQUIRED,
            }
            if cert_reqs not in CERT_REQS:
                raise RedisError(
                    "Invalid SSL Certificate Requirements Flag: %s" % cert_reqs
                )
            self.cert_reqs = CERT_REQS[cert_reqs]
        self.ca_certs = ca_certs
        self.context = None

    def get(self):
        if not self.keyfile:
            self.context = ssl.create_default_context(cafile=self.ca_certs)
        else:
            self.context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
            self.context.verify_mode = self.cert_reqs
            self.context.load_cert_chain(certfile=self.certfile, keyfile=self.keyfile)
            self.context.load_verify_locations(self.ca_certs)
        return self.context


class BaseConnection:
    description = "BaseConnection"

    def __init__(
        self,
        retry_on_timeout=False,
        stream_timeout=None,
        parser_class=DefaultParser,
        reader_read_size=65535,
        encoding="utf-8",
        decode_responses=False,
        *,
        loop=None
    ):
        self._parser = parser_class(reader_read_size)
        self._stream_timeout = stream_timeout
        self._reader = None
        self._writer = None
        self.password = ""
        self.db = ""
        self.pid = os.getpid()
        self.retry_on_timeout = retry_on_timeout
        self._description_args = dict()
        self._connect_callbacks = list()
        self.encoding = encoding
        self.decode_responses = decode_responses
        self.loop = loop
        # flag to show if a connection is waiting for response
        self.awaiting_response = False
        self.last_active_at = time.time()

    def __repr__(self):
        return self.description.format(**self._description_args)

    def __del__(self):
        try:
            self.disconnect()
        except Exception:
            pass

    @property
    def is_connected(self):
        return bool(self._reader and self._writer)

    def register_connect_callback(self, callback):
        self._connect_callbacks.append(callback)

    def clear_connect_callbacks(self):
        self._connect_callbacks = list()

    async def can_read(self):
        """Checks for data that can be read"""
        if not self.is_connected:
            await self.connect()
        return self._parser.can_read()

    async def connect(self):
        try:
            await self._connect()
        except asyncio.CancelledError:
            raise
        except Exception:
            raise ConnectionError()
        # run any user callbacks. right now the only internal callback
        # is for pubsub channel/pattern resubscription
        for callback in self._connect_callbacks:
            task = callback(self)
            # typing.Awaitable is not available in Python3.5
            # so use inspect.isawaitable instead
            # according to issue https://github.com/alisaifee/coredis/issues/77
            if inspect.isawaitable(task):
                await task

    async def _connect(self):
        raise NotImplementedError

    async def on_connect(self):
        self._parser.on_connect(self)

        # if a password is specified, authenticate
        if self.password:
            await self.send_command("AUTH", self.password)
            if nativestr(await self.read_response()) != "OK":
                raise ConnectionError("Invalid Password")

        # if a database is specified, switch to it
        if self.db:
            await self.send_command("SELECT", self.db)
            if nativestr(await self.read_response()) != "OK":
                raise ConnectionError("Invalid Database")
        self.last_active_at = time.time()

    async def read_response(self):
        try:
            response = await exec_with_timeout(
                self._parser.read_response(), self._stream_timeout, loop=self.loop
            )
            self.last_active_at = time.time()
        except TimeoutError:
            self.disconnect()
            raise
        if isinstance(response, RedisError):
            raise response
        self.awaiting_response = False
        return response

    async def send_packed_command(self, command):
        """Sends an already packed command to the Redis server"""
        if not self._writer:
            await self.connect()
        try:
            if isinstance(command, str):
                command = [command]
            self._writer.writelines(command)
        except TimeoutError:
            self.disconnect()
            raise TimeoutError("Timeout writing to socket")
        except Exception:
            e = sys.exc_info()[1]
            self.disconnect()
            if len(e.args) == 1:
                errno, errmsg = "UNKNOWN", e.args[0]
            else:
                errno = e.args[0]
                errmsg = e.args[1]
            raise ConnectionError(
                "Error %s while writing to socket. %s." % (errno, errmsg)
            )
        except Exception:
            self.disconnect()
            raise

    async def send_command(self, *args):
        if not self.is_connected:
            await self.connect()
        await self.send_packed_command(self.pack_command(*args))
        self.awaiting_response = True
        self.last_active_at = time.time()

    def encode(self, value):
        """Returns a bytestring representation of the value"""
        if isinstance(value, bytes):
            return value
        elif isinstance(value, int):
            value = b(str(value))
        elif isinstance(value, float):
            value = b(repr(value))
        elif not isinstance(value, str):
            value = str(value)
        if isinstance(value, str):
            value = value.encode(self.encoding)
        return value

    def disconnect(self):
        """Disconnects from the Redis server"""
        self._parser.on_disconnect()
        try:
            self._writer.close()
        except Exception:
            pass
        self._reader = None
        self._writer = None

    def pack_command(self, *args):
        "Pack a series of arguments into the Redis protocol"
        output = []
        # the client might have included 1 or more literal arguments in
        # the command name, e.g., 'CONFIG GET'. The Redis server expects these
        # arguments to be sent separately, so split the first argument
        # manually. All of these arguements get wrapped in the Token class
        # to prevent them from being encoded.
        command = args[0]
        if " " in command:
            args = tuple([b(s) for s in command.split()]) + args[1:]
        else:
            args = (b(command),) + args[1:]

        buff = SYM_EMPTY.join((SYM_STAR, b(str(len(args))), SYM_CRLF))
        for arg in map(self.encode, args):
            # to avoid large string mallocs, chunk the command into the
            # output list if we're sending large values
            if len(buff) > 6000 or len(arg) > 6000:
                buff = SYM_EMPTY.join((buff, SYM_DOLLAR, b(str(len(arg))), SYM_CRLF))
                output.append(buff)
                output.append(b(arg))
                buff = SYM_CRLF
            else:
                buff = SYM_EMPTY.join(
                    (buff, SYM_DOLLAR, b(str(len(arg))), SYM_CRLF, b(arg), SYM_CRLF)
                )
        output.append(buff)
        return output

    def pack_commands(self, commands):
        "Pack multiple commands into the Redis protocol"
        output = []
        pieces = []
        buffer_length = 0

        for cmd in commands:
            for chunk in self.pack_command(*cmd):
                pieces.append(chunk)
                buffer_length += len(chunk)

            if buffer_length > 6000:
                output.append(SYM_EMPTY.join(pieces))
                buffer_length = 0
                pieces = []

        if pieces:
            output.append(SYM_EMPTY.join(pieces))
        return output


class Connection(BaseConnection):
    description = "Connection<host={host},port={port},db={db}>"

    def __init__(
        self,
        host="127.0.0.1",
        port=6379,
        password=None,
        db=0,
        retry_on_timeout=False,
        stream_timeout=None,
        connect_timeout=None,
        ssl_context=None,
        parser_class=DefaultParser,
        reader_read_size=65535,
        encoding="utf-8",
        decode_responses=False,
        socket_keepalive=None,
        socket_keepalive_options=None,
        *,
        loop=None
    ):
        super(Connection, self).__init__(
            retry_on_timeout,
            stream_timeout,
            parser_class,
            reader_read_size,
            encoding,
            decode_responses,
            loop=loop,
        )
        self.host = host
        self.port = port
        self.password = password
        self.db = db
        self.ssl_context = ssl_context
        self._connect_timeout = connect_timeout
        self._description_args = {"host": self.host, "port": self.port, "db": self.db}
        self.socket_keepalive = socket_keepalive
        self.socket_keepalive_options = socket_keepalive_options or {}

    async def _connect(self):
        if LOOP_DEPRECATED:
            connection = asyncio.open_connection(
                host=self.host, port=self.port, ssl=self.ssl_context
            )
        else:
            connection = asyncio.open_connection(
                host=self.host, port=self.port, ssl=self.ssl_context, loop=self.loop
            )
        reader, writer = await exec_with_timeout(
            connection, self._connect_timeout, loop=self.loop
        )
        self._reader = reader
        self._writer = writer
        sock = writer.transport.get_extra_info("socket")
        if sock is not None:
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            try:
                # TCP_KEEPALIVE
                if self.socket_keepalive:
                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                    for k, v in self.socket_keepalive_options.items():
                        sock.setsockopt(socket.SOL_TCP, k, v)
            except (socket.error, TypeError):
                # `socket_keepalive_options` might contain invalid options
                # causing an error. Do not leave the connection open.
                writer.close()
                raise
        await self.on_connect()


class UnixDomainSocketConnection(BaseConnection):
    description = "UnixDomainSocketConnection<path={path},db={db}>"

    def __init__(
        self,
        path="",
        password=None,
        db=0,
        retry_on_timeout=False,
        stream_timeout=None,
        connect_timeout=None,
        ssl_context=None,
        parser_class=DefaultParser,
        reader_read_size=65535,
        encoding="utf-8",
        decode_responses=False,
        *,
        loop=None
    ):
        super(UnixDomainSocketConnection, self).__init__(
            retry_on_timeout,
            stream_timeout,
            parser_class,
            reader_read_size,
            encoding,
            decode_responses,
            loop=loop,
        )
        self.path = path
        self.db = db
        self.password = password
        self.ssl_context = ssl_context
        self._connect_timeout = connect_timeout
        self._description_args = {"path": self.path, "db": self.db}

    async def _connect(self):
        if LOOP_DEPRECATED:
            connection = asyncio.open_unix_connection(
                path=self.path, ssl=self.ssl_context
            )
        else:
            connection = asyncio.open_unix_connection(
                path=self.path, ssl=self.ssl_context, loop=self.loop
            )
        reader, writer = await exec_with_timeout(
            connection, self._connect_timeout, loop=self.loop
        )
        self._reader = reader
        self._writer = writer
        await self.on_connect()


class ClusterConnection(Connection):
    "Manages TCP communication to and from a Redis server"
    description = "ClusterConnection<host={host},port={port}>"

    def __init__(self, *args, **kwargs):
        self.readonly = kwargs.pop("readonly", False)
        super(ClusterConnection, self).__init__(*args, **kwargs)

    async def on_connect(self):
        """
        Initialize the connection, authenticate and select a database and send READONLY if it is
        set during object initialization.
        """
        if self.db:
            warnings.warn("SELECT DB is not allowed in cluster mode")
            self.db = ""
        await super(ClusterConnection, self).on_connect()
        if self.readonly:
            await self.send_command("READONLY")
            if nativestr(await self.read_response()) != "OK":
                raise ConnectionError("READONLY command failed")



#!/usr/bin/python
# -*- coding: utf-8 -*-
import socket

import pytest

from coredis import Connection


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_connect_tcp(event_loop):
    conn = Connection(loop=event_loop)
    assert conn.host == "127.0.0.1"
    assert conn.port == 6379
    assert str(conn) == "Connection<host=127.0.0.1,port=6379,db=0>"
    await conn.send_command("PING")
    res = await conn.read_response()
    assert res == b"PONG"
    assert (conn._reader is not None) and (conn._writer is not None)
    conn.disconnect()
    assert (conn._reader is None) and (conn._writer is None)


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_connect_tcp_keepalive_options(event_loop):
    conn = Connection(
        loop=event_loop,
        socket_keepalive=True,
        socket_keepalive_options={
            socket.TCP_KEEPIDLE: 1,
            socket.TCP_KEEPINTVL: 1,
            socket.TCP_KEEPCNT: 3,
        },
    )
    await conn._connect()
    sock = conn._writer.transport.get_extra_info("socket")
    assert sock.getsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE) == 1
    for k, v in (
        (socket.TCP_KEEPIDLE, 1),
        (socket.TCP_KEEPINTVL, 1),
        (socket.TCP_KEEPCNT, 3),
    ):
        assert sock.getsockopt(socket.SOL_TCP, k) == v
    conn.disconnect()


@pytest.mark.parametrize("option", ["UNKNOWN", 999])
@pytest.mark.asyncio(forbid_global_loop=True)
async def test_connect_tcp_wrong_socket_opt_raises(event_loop, option):
    conn = Connection(
        loop=event_loop, socket_keepalive=True, socket_keepalive_options={option: 1}
    )
    with pytest.raises((socket.error, TypeError)):
        await conn._connect()
    # verify that the connection isn't left open
    assert conn._writer.transport.is_closing()


# only test during dev
# @pytest.mark.asyncio(forbid_global_loop=True)
# async def test_connect_unix_socket(event_loop):
#     # to run this test case you should change your redis configuration
#     # unixsocket /var/run/redis/redis.sock
#     # unixsocketperm 777
#     path = '/var/run/redis/redis.sock'
#     conn = UnixDomainSocketConnection(path, event_loop)
#     await conn.connect()
#     assert conn.path == path
#     assert str(conn) == 'UnixDomainSocketConnection<path={},db=0>'.format(path)
#     await conn.send_command('PING')
#     res = await conn.read_response()
#     assert res == b'PONG'
#     assert (conn._reader is not None) and (conn._writer is not None)
#     conn.disconnect()
#     assert (conn._reader is None) and (conn._writer is None)



from coredis.exceptions import DataError
from coredis.utils import (
    b,
    dict_merge,
    first_key,
    iteritems,
    list_or_args,
    pairs_to_dict,
    string_keys_to_dict,
)


def parse_hscan(response, **options):
    cursor, r = response

    return int(cursor), r and pairs_to_dict(r) or {}


class HashCommandMixin:

    RESPONSE_CALLBACKS = dict_merge(
        string_keys_to_dict("HDEL HLEN", int),
        string_keys_to_dict("HEXISTS HMSET", bool),
        {
            "HGETALL": lambda r: r and pairs_to_dict(r) or {},
            "HINCRBYFLOAT": float,
            "HSCAN": parse_hscan,
        },
    )

    async def hdel(self, name, *keys):
        """Deletes ``keys`` from hash ``name``"""

        return await self.execute_command("HDEL", name, *keys)

    async def hexists(self, name, key):
        """
        Returns a boolean indicating if ``key`` exists within hash ``name``
        """

        return await self.execute_command("HEXISTS", name, key)

    async def hget(self, name, key):
        """Returns the value of ``key`` within the hash ``name``"""

        return await self.execute_command("HGET", name, key)

    async def hgetall(self, name):
        """Returns a Python dict of the hash's name/value pairs"""

        return await self.execute_command("HGETALL", name)

    async def hincrby(self, name, key, amount=1):
        """Increments the value of ``key`` in hash ``name`` by ``amount``"""

        return await self.execute_command("HINCRBY", name, key, amount)

    async def hincrbyfloat(self, name, key, amount=1.0):
        """
        Increments the value of ``key`` in hash ``name`` by floating
        ``amount``
        """

        return await self.execute_command("HINCRBYFLOAT", name, key, amount)

    async def hkeys(self, name):
        """Returns the list of keys within hash ``name``"""

        return await self.execute_command("HKEYS", name)

    async def hlen(self, name):
        """Returns the number of elements in hash ``name``"""

        return await self.execute_command("HLEN", name)

    async def hset(self, name, key, value):
        """
        Sets ``key`` to ``value`` within hash ``name``
        Returns 1 if HSET created a new field, otherwise 0
        """

        return await self.execute_command("HSET", name, key, value)

    async def hsetnx(self, name, key, value):
        """
        Sets ``key`` to ``value`` within hash ``name`` if ``key`` does not
        exist.  Returns 1 if HSETNX created a field, otherwise 0.
        """

        return await self.execute_command("HSETNX", name, key, value)

    async def hmset(self, name, mapping):
        """
        Sets key to value within hash ``name`` for each corresponding
        key and value from the ``mapping`` dict.
        """

        if not mapping:
            raise DataError("'hmset' with 'mapping' of length 0")
        items = []

        for pair in iteritems(mapping):
            items.extend(pair)

        return await self.execute_command("HMSET", name, *items)

    async def hmget(self, name, keys, *args):
        """Returns a list of values ordered identically to ``keys``"""
        args = list_or_args(keys, args)

        return await self.execute_command("HMGET", name, *args)

    async def hvals(self, name):
        """Returns the list of values within hash ``name``"""

        return await self.execute_command("HVALS", name)

    async def hscan(self, name, cursor=0, match=None, count=None):
        """
        Incrementallys return key/value slices in a hash. Also returns a
        cursor pointing to the scan position.

        :param match: allows for filtering the keys by pattern
        :param count: allows for hint the minimum number of returns
        """
        pieces = [name, cursor]

        if match is not None:
            pieces.extend([b("MATCH"), match])

        if count is not None:
            pieces.extend([b("COUNT"), count])

        return await self.execute_command("HSCAN", *pieces)

    async def hstrlen(self, name, key):
        """
        Returns the string length of the value associated
        with field in the hash stored at key.
        If the key or the field do not exist, 0 is returned.
        """

        return await self.execute_command("HSTRLEN", name, key)

    async def hrandfield(self, key, count=None, withvalues=False):
        """
        Return a random field from the hash value stored at key.

        :param count: if the argument is positive, return an array of distinct fields.
         If called with a negative count, the behavior changes and the command
         is allowed to return the same field multiple times. In this case,
         the number of returned fields is the absolute value of the
         specified count.
        :param withvalues: The optional WITHVALUES modifier changes the reply so it
         includes the respective values of the randomly selected hash fields.

        .. versionadded:: 2.1.0
        """
        params = []

        if count is not None:
            params.append(count)

        if withvalues:
            params.append("WITHVALUES")

        return await self.execute_command("HRANDFIELD", key, *params)


class ClusterHashCommandMixin(HashCommandMixin):

    RESULT_CALLBACKS = {"HSCAN": first_key}



#!/usr/bin/python
# -*- coding: utf-8 -*-
import asyncio
from unittest.mock import Mock

import pytest
from packaging import version

import coredis

_REDIS_VERSIONS = {}


async def get_version(**kwargs):
    params = {"host": "localhost", "port": 6379, "db": 0}
    params.update(kwargs)
    key = "%s:%s" % (params["host"], params["port"])

    if key not in _REDIS_VERSIONS:
        client = coredis.StrictRedis(**params)
        _REDIS_VERSIONS[key] = (await client.info())["redis_version"]
        client.connection_pool.disconnect()

    return _REDIS_VERSIONS[key]


def skip_if_server_version_lt(min_version):
    loop = asyncio.get_event_loop()
    cur = version.parse(loop.run_until_complete(get_version()))
    check = cur < version.parse(min_version)

    return pytest.mark.skipif(check, reason="")


@pytest.fixture()
async def r(event_loop):
    client = coredis.StrictRedis(loop=event_loop)

    await client.config_set("maxmemory-policy", "noeviction")
    await client.flushdb()
    return client


class AsyncMock(Mock):
    def __init__(self, *args, **kwargs):
        super(AsyncMock, self).__init__(*args, **kwargs)

    def __await__(self):
        future = asyncio.Future(loop=self.loop)
        future.set_result(self)
        result = yield from future

        return result

    @staticmethod
    def pack_response(response, *, loop):
        future = asyncio.Future(loop=loop)
        future.set_result(response)

        return future


def _gen_mock_resp(r, response, *, loop):
    mock_connection_pool = AsyncMock(loop=loop)
    connection = AsyncMock(loop=loop)
    connection.read_response.return_value = AsyncMock.pack_response(response, loop=loop)
    mock_connection_pool.get_connection.return_value = connection
    r.connection_pool = mock_connection_pool

    return r


@pytest.fixture()
def mock_resp_role(event_loop):
    r = coredis.StrictRedis(loop=event_loop)
    response = [b"master", 169, [[b"172.17.0.2", b"7004", b"169"]]]

    return _gen_mock_resp(r, response, loop=event_loop)



from coredis.client import StrictRedis, StrictRedisCluster
from coredis.commands.strings import BitFieldOperation
from coredis.connection import ClusterConnection, Connection, UnixDomainSocketConnection
from coredis.exceptions import (
    AskError,
    BusyLoadingError,
    CacheError,
    ClusterCrossSlotError,
    ClusterDownError,
    ClusterError,
    ClusterTransactionError,
    CompressError,
    ConnectionError,
    DataError,
    ExecAbortError,
    InvalidResponse,
    LockError,
    MovedError,
    NoScriptError,
    PubSubError,
    ReadOnlyError,
    RedisClusterException,
    RedisError,
    ResponseError,
    SerializeError,
    TimeoutError,
    TryAgainError,
    WatchError,
)
from coredis.pool import ClusterConnectionPool, ConnectionPool

from . import _version

__all__ = [
    "StrictRedis",
    "StrictRedisCluster",
    "BitFieldOperation",
    "Connection",
    "UnixDomainSocketConnection",
    "ClusterConnection",
    "ConnectionPool",
    "ClusterConnectionPool",
    "AskError",
    "BusyLoadingError",
    "CacheError",
    "ClusterCrossSlotError",
    "ClusterDownError",
    "ClusterError",
    "ClusterTransactionError",
    "CompressError",
    "ConnectionError",
    "DataError",
    "ExecAbortError",
    "InvalidResponse",
    "LockError",
    "MovedError",
    "NoScriptError",
    "PubSubError",
    "ReadOnlyError",
    "RedisClusterException",
    "RedisError",
    "ResponseError",
    "SerializeError",
    "TimeoutError",
    "TryAgainError",
    "WatchError",
]


__version__ = _version.get_versions()["version"]



from __future__ import with_statement

import pytest

from coredis.exceptions import ResponseError, WatchError
from coredis.utils import b


class TestPipeline:
    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_pipeline(self, r):
        async with await r.pipeline() as pipe:
            await pipe.set("a", "a1")
            await pipe.get("a")
            await pipe.zadd("z", z1=1)
            await pipe.zadd("z", z2=4)
            await pipe.zincrby("z", "z1")
            await pipe.zrange("z", 0, 5, withscores=True)
            assert await pipe.execute() == [
                True,
                b("a1"),
                1,
                1,
                2.0,
                [(b("z1"), 2.0), (b("z2"), 4)],
            ]

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_pipeline_length(self, r):
        async with await r.pipeline() as pipe:
            # Initially empty.
            assert len(pipe) == 0
            assert not pipe

            # Fill 'er up!
            await pipe.set("a", "a1")
            await pipe.set("b", "b1")
            await pipe.set("c", "c1")
            assert len(pipe) == 3
            assert pipe

            # Execute calls reset(), so empty once again.
            await pipe.execute()
            assert len(pipe) == 0
            assert not pipe

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_pipeline_no_transaction(self, r):
        async with await r.pipeline(transaction=False) as pipe:
            await pipe.set("a", "a1")
            await pipe.set("b", "b1")
            await pipe.set("c", "c1")
            assert await pipe.execute() == [True, True, True]
            assert await r.get("a") == b("a1")
            assert await r.get("b") == b("b1")
            assert await r.get("c") == b("c1")

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_pipeline_no_transaction_watch(self, r):
        await r.set("a", 0)

        async with await r.pipeline(transaction=False) as pipe:
            await pipe.watch("a")
            a = await pipe.get("a")

            pipe.multi()
            await pipe.set("a", int(a) + 1)
            assert await pipe.execute() == [True]

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_pipeline_no_transaction_watch_failure(self, r):
        await r.set("a", 0)

        async with await r.pipeline(transaction=False) as pipe:
            await pipe.watch("a")
            a = await pipe.get("a")

            await r.set("a", "bad")

            pipe.multi()
            await pipe.set("a", int(a) + 1)

            with pytest.raises(WatchError):
                await pipe.execute()

            assert await r.get("a") == b("bad")

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_exec_error_in_response(self, r):
        """
        an invalid pipeline command at exec time adds the exception instance
        to the list of returned values
        """
        await r.set("c", "a")
        async with await r.pipeline() as pipe:
            await pipe.set("a", 1)
            await pipe.set("b", 2)
            await pipe.lpush("c", 3)
            await pipe.set("d", 4)
            result = await pipe.execute(raise_on_error=False)

            assert result[0]
            assert await r.get("a") == b("1")
            assert result[1]
            assert await r.get("b") == b("2")

            # we can't lpush to a key that's a string value, so this should
            # be a ResponseError exception
            assert isinstance(result[2], ResponseError)
            assert await r.get("c") == b("a")

            # since this isn't a transaction, the other commands after the
            # error are still executed
            assert result[3]
            assert await r.get("d") == b("4")

            # make sure the pipe was restored to a working state
            await pipe.set("z", "zzz")
            assert await pipe.execute() == [True]
            assert await r.get("z") == b("zzz")

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_exec_error_raised(self, r):
        await r.set("c", "a")
        async with await r.pipeline() as pipe:
            await pipe.set("a", 1)
            await pipe.set("b", 2)
            await pipe.lpush("c", 3)
            await pipe.set("d", 4)
            with pytest.raises(ResponseError):
                await pipe.execute()

            # make sure the pipe was restored to a working state
            await pipe.set("z", "zzz")
            assert await pipe.execute() == [True]
            assert await r.get("z") == b("zzz")

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_parse_error_raised(self, r):
        async with await r.pipeline() as pipe:
            # the zrem is invalid because we don't pass any keys to it
            await pipe.set("a", 1)
            await pipe.zrem("b")
            await pipe.set("b", 2)
            with pytest.raises(ResponseError):
                await pipe.execute()

            # make sure the pipe was restored to a working state
            await pipe.set("z", "zzz")
            assert await pipe.execute() == [True]
            assert await r.get("z") == b("zzz")

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_watch_succeed(self, r):
        await r.set("a", 1)
        await r.set("b", 2)

        async with await r.pipeline() as pipe:
            await pipe.watch("a", "b")
            assert pipe.watching
            a_value = await pipe.get("a")
            b_value = await pipe.get("b")
            assert a_value == b("1")
            assert b_value == b("2")
            pipe.multi()

            await pipe.set("c", 3)
            assert await pipe.execute() == [True]
            assert not pipe.watching

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_watch_failure(self, r):
        await r.set("a", 1)
        await r.set("b", 2)

        async with await r.pipeline() as pipe:
            await pipe.watch("a", "b")
            await r.set("b", 3)
            pipe.multi()
            await pipe.get("a")
            with pytest.raises(WatchError):
                await pipe.execute()

            assert not pipe.watching

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_unwatch(self, r):
        await r.set("a", 1)
        await r.set("b", 2)

        async with await r.pipeline() as pipe:
            await pipe.watch("a", "b")
            await r.set("b", 3)
            await pipe.unwatch()
            assert not pipe.watching
            await pipe.get("a")
            assert await pipe.execute() == [b("1")]

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_transaction_callable(self, r):
        await r.set("a", 1)
        await r.set("b", 2)
        has_run = []

        async def my_transaction(pipe):
            a_value = await pipe.get("a")
            assert a_value in (b("1"), b("2"))
            b_value = await pipe.get("b")
            assert b_value == b("2")

            # silly run-once code... incr's "a" so WatchError should be raised
            # forcing this all to run again. this should incr "a" once to "2"
            if not has_run:
                await r.incr("a")
                has_run.append("it has")

            pipe.multi()
            await pipe.set("c", int(a_value) + int(b_value))

        result = await r.transaction(my_transaction, "a", "b", watch_delay=0.01)
        assert result == [True]
        assert await r.get("c") == b("4")

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_exec_error_in_no_transaction_pipeline(self, r):
        await r.set("a", 1)
        async with await r.pipeline(transaction=False) as pipe:
            await pipe.llen("a")
            await pipe.expire("a", 100)

            with pytest.raises(ResponseError):
                await pipe.execute()

        assert await r.get("a") == b("1")

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_exec_error_in_no_transaction_pipeline_unicode_command(self, r):
        key = chr(11) + "abcd" + chr(23)
        await r.set(key, 1)
        async with await r.pipeline(transaction=False) as pipe:
            await pipe.llen(key)
            await pipe.expire(key, 100)

            with pytest.raises(ResponseError):
                await pipe.execute()

        assert await r.get(key) == b("1")



from coredis.exceptions import DataError, RedisError
from coredis.utils import (
    b,
    dict_merge,
    first_key,
    int_or_none,
    iteritems,
    iterkeys,
    itervalues,
    list_or_args,
    string_keys_to_dict,
)

VALID_ZADD_OPTIONS = {"NX", "XX", "CH", "INCR"}


def float_or_none(response):
    if response is None:
        return None

    return float(response)


def zset_score_pairs(response, **options):
    """
    If ``withscores`` is specified in the options, return the response as
    a list of (value, score) pairs
    """

    if not response or not options.get("withscores"):
        return response
    score_cast_func = options.get("score_cast_func", float)
    it = iter(response)

    return list(zip(it, map(score_cast_func, it)))


def parse_zmscore(response, **options):
    return [float(score) if score is not None else None for score in response]


def parse_zscan(response, **options):
    score_cast_func = options.get("score_cast_func", float)
    cursor, r = response
    it = iter(r)

    return int(cursor), list(zip(it, map(score_cast_func, it)))


class SortedSetCommandMixin:

    RESPONSE_CALLBACKS = dict_merge(
        string_keys_to_dict(
            "ZADD ZCARD ZLEXCOUNT "
            "ZREM ZREMRANGEBYLEX "
            "ZREMRANGEBYRANK "
            "ZREMRANGEBYSCORE",
            int,
        ),
        string_keys_to_dict("ZSCORE ZINCRBY", float_or_none),
        string_keys_to_dict(
            "ZPOPMAX ZPOPMIN ZINTER ZDIFF ZUNION ZRANGE ZRANGEBYSCORE "
            "ZREVRANGE ZREVRANGEBYSCORE",
            zset_score_pairs,
        ),
        string_keys_to_dict("ZRANK ZREVRANK", int_or_none),
        string_keys_to_dict(
            "BZPOPMIN BZPOPMAX", lambda r: r and (r[0], r[1], float(r[2])) or None
        ),
        {"ZSCAN": parse_zscan},
        {"ZMSCORE": parse_zmscore},
    )

    async def zadd(self, name, *args, **kwargs):
        """
        Set any number of score, element-name pairs to the key ``name``. Pairs
        can be specified in two ways:

        As ``*args``, in the form of: score1, name1, score2, name2, ...
        or as ``**kwargs``, in the form of: name1=score1, name2=score2, ...

        The following example would add four values to the 'my-key' key:
        redis.zadd('my-key', 1.1, 'name1', 2.2, 'name2', name3=3.3, name4=4.4)
        """
        pieces = []

        if args:
            if len(args) % 2 != 0:
                raise RedisError(
                    "ZADD requires an equal number of " "values and scores"
                )
            pieces.extend(args)

        for pair in iteritems(kwargs):
            pieces.append(pair[1])
            pieces.append(pair[0])

        return await self.execute_command("ZADD", name, *pieces)

    async def zaddoption(self, name, option=None, *args, **kwargs):
        """
        Differs from zadd in that you can set either 'XX' or 'NX' option as
        described here: https://redis.io/commands/zadd. Only for Redis 3.0.2 or
        later.

        The following example would add four values to the 'my-key' key:
        redis.zaddoption('my-key', 'XX', 1.1, 'name1', 2.2, 'name2', name3=3.3, name4=4.4)
        redis.zaddoption('my-key', 'NX CH', name1=2.2)
        """

        if not option:
            raise RedisError("ZADDOPTION must take options")
        options = set(opt.upper() for opt in option.split())

        if options - VALID_ZADD_OPTIONS:
            raise RedisError("ZADD only takes XX, NX, CH, or INCR")

        if "NX" in options and "XX" in options:
            raise RedisError("ZADD only takes one of XX or NX")
        pieces = list(options)
        members = []

        if args:
            if len(args) % 2 != 0:
                raise RedisError(
                    "ZADD requires an equal number of " "values and scores"
                )
            members.extend(args)

        for pair in iteritems(kwargs):
            members.append(pair[1])
            members.append(pair[0])

        if "INCR" in options and len(members) != 2:
            raise RedisError("ZADD with INCR only takes one score-name pair")

        return await self.execute_command("ZADD", name, *pieces, *members)

    async def zcard(self, name):
        """Returns the number of elements in the sorted set ``name``"""

        return await self.execute_command("ZCARD", name)

    async def zcount(self, name, min, max):
        """
        Returns the number of elements in the sorted set at key ``name`` with
        a score between ``min`` and ``max``.
        """

        return await self.execute_command("ZCOUNT", name, min, max)

    async def zdiff(self, keys, withscores=False):
        """
        Returns the difference between the first and all successive input
        sorted sets provided in ``keys``.

        .. versionadded:: 2.1.0
        """
        pieces = [len(keys), *keys]

        if withscores:
            pieces.append("WITHSCORES")

        return await self.execute_command("ZDIFF", *pieces)

    async def zdiffstore(self, dest, keys):
        """
        Computes the difference between the first and all successive input
        sorted sets provided in ``keys`` and stores the result in ``dest``.

        .. versionadded:: 2.1.0
        """
        pieces = [len(keys), *keys]

        return await self.execute_command("ZDIFFSTORE", dest, *pieces)

    async def zincrby(self, name, value, amount=1):
        """
        Increments the score of ``value`` in sorted set ``name`` by ``amount``
        """

        return await self.execute_command("ZINCRBY", name, amount, value)

    async def zinter(self, keys, aggregate=None, withscores=False):
        """
        Return the intersect of multiple sorted sets specified by ``keys``.
        With the ``aggregate`` option, it is possible to specify how the
        results of the union are aggregated. This option defaults to SUM,
        where the score of an element is summed across the inputs where it
        exists. When this option is set to either MIN or MAX, the resulting
        set will contain the minimum or maximum score of an element across
        the inputs where it exists.

        .. versionadded:: 2.1.0
        """

        return await self._zaggregate(
            "ZINTER", None, keys, aggregate, withscores=withscores
        )

    async def zinterstore(self, dest, keys, aggregate=None):
        """
        Intersects multiple sorted sets specified by ``keys`` into
        a new sorted set, ``dest``. Scores in the destination will be
        aggregated based on the ``aggregate``, or SUM if none is provided.

        .. versionadded:: 2.1.0
        """

        return await self._zaggregate("ZINTERSTORE", dest, keys, aggregate)

    async def zlexcount(self, name, min, max):
        """
        Returns the number of items in the sorted set ``name`` between the
        lexicographical range ``min`` and ``max``.
        """

        return await self.execute_command("ZLEXCOUNT", name, min, max)

    async def zpopmax(self, name, count=None):
        """
        Remove and return up to ``count`` members with the highest scores
        from the sorted set ``name``.

        .. versionadded:: 2.1.0
        """
        args = (count is not None) and [count] or []
        options = {"withscores": True}

        return await self.execute_command("ZPOPMAX", name, *args, **options)

    async def zpopmin(self, name, count=None):
        """
        Remove and return up to ``count`` members with the lowest scores
        from the sorted set ``name``.

        .. versionadded:: 2.1.0
        """
        args = (count is not None) and [count] or []
        options = {"withscores": True}

        return await self.execute_command("ZPOPMIN", name, *args, **options)

    async def zrandmember(self, key, count=None, withscores=False):
        """
        Return a random element from the sorted set value stored at key.
        ``count`` if the argument is positive, return an array of distinct
        fields. If called with a negative count, the behavior changes and
        the command is allowed to return the same field multiple times.
        In this case, the number of returned fields is the absolute value
        of the specified count.
        ``withscores`` The optional WITHSCORES modifier changes the reply so it
        includes the respective scores of the randomly selected elements from
        the sorted set.

        .. versionadded:: 2.1.0
        """
        params = []

        if count is not None:
            params.append(count)

        if withscores:
            params.append("WITHSCORES")

        return await self.execute_command("ZRANDMEMBER", key, *params)

    async def bzpopmax(self, keys, timeout=0):
        """
        ZPOPMAX a value off of the first non-empty sorted set
        named in the ``keys`` list.
        If none of the sorted sets in ``keys`` has a value to ZPOPMAX,
        then block for ``timeout`` seconds, or until a member gets added
        to one of the sorted sets.
        If timeout is 0, then block indefinitely.

        .. versionadded:: 2.1.0
        """

        if timeout is None:
            timeout = 0
        keys = list_or_args(keys, None)
        keys.append(timeout)

        return await self.execute_command("BZPOPMAX", *keys)

    async def bzpopmin(self, keys, timeout=0):
        """
        ZPOPMIN a value off of the first non-empty sorted set
        named in the ``keys`` list.
        If none of the sorted sets in ``keys`` has a value to ZPOPMIN,
        then block for ``timeout`` seconds, or until a member gets added
        to one of the sorted sets.
        If timeout is 0, then block indefinitely.

        .. versionadded:: 2.1.0
        """

        if timeout is None:
            timeout = 0
        keys = list_or_args(keys, None)
        keys.append(timeout)

        return await self.execute_command("BZPOPMIN", *keys)

    async def _zrange(
        self,
        command,
        dest,
        name,
        start,
        end,
        desc=False,
        byscore=False,
        bylex=False,
        withscores=False,
        score_cast_func=float,
        offset=None,
        num=None,
    ):
        if byscore and bylex:
            raise DataError(
                "``byscore`` and ``bylex`` can not be " "specified together."
            )

        if (offset is not None and num is None) or (num is not None and offset is None):
            raise DataError("``offset`` and ``num`` must both be specified.")

        if bylex and withscores:
            raise DataError(
                "``withscores`` not supported in combination " "with ``bylex``."
            )
        pieces = [command]

        if dest:
            pieces.append(dest)
        pieces.extend([name, start, end])

        if byscore:
            pieces.append("BYSCORE")

        if bylex:
            pieces.append("BYLEX")

        if desc:
            pieces.append("REV")

        if offset is not None and num is not None:
            pieces.extend(["LIMIT", offset, num])

        if withscores:
            pieces.append("WITHSCORES")
        options = {"withscores": withscores, "score_cast_func": score_cast_func}

        return await self.execute_command(*pieces, **options)

    async def zrange(
        self,
        name,
        start,
        end,
        desc=False,
        withscores=False,
        score_cast_func=float,
        byscore=False,
        bylex=False,
        offset=None,
        num=None,
    ):
        """
        Return a range of values from sorted set ``name`` between
        ``start`` and ``end`` sorted in ascending order.
        ``start`` and ``end`` can be negative, indicating the end of the range.
        ``desc`` a boolean indicating whether to sort the results in reversed
        order.
        ``withscores`` indicates to return the scores along with the values.
        The return type is a list of (value, score) pairs.
        ``score_cast_func`` a callable used to cast the score return value.
        ``byscore`` when set to True, returns the range of elements from the
        sorted set having scores equal or between ``start`` and ``end``.
        ``bylex`` when set to True, returns the range of elements from the
        sorted set between the ``start`` and ``end`` lexicographical closed
        range intervals.
        Valid ``start`` and ``end`` must start with ( or [, in order to specify
        whether the range interval is exclusive or inclusive, respectively.
        ``offset`` and ``num`` are specified, then return a slice of the range.
        Can't be provided when using ``bylex``.
        For more information check https://redis.io/commands/zrange
        """
        # Need to support ``desc`` also when using old redis version
        # because it was supported in 3.5.3 (of redis-py)

        if not byscore and not bylex and (offset is None and num is None) and desc:
            return self.zrevrange(name, start, end, withscores, score_cast_func)

        return await self._zrange(
            "ZRANGE",
            None,
            name,
            start,
            end,
            desc,
            byscore,
            bylex,
            withscores,
            score_cast_func,
            offset,
            num,
        )

    async def zrangebylex(self, name, min, max, start=None, num=None):
        """
        Returns the lexicographical range of values from sorted set ``name``
        between ``min`` and ``max``.

        If ``start`` and ``num`` are specified, then return a slice of the
        range.
        """

        if (start is not None and num is None) or (num is not None and start is None):
            raise RedisError("``start`` and ``num`` must both be specified")
        pieces = ["ZRANGEBYLEX", name, min, max]

        if start is not None and num is not None:
            pieces.extend([b("LIMIT"), start, num])

        return await self.execute_command(*pieces)

    async def zrevrangebylex(self, name, max, min, start=None, num=None):
        """
        Returns the reversed lexicographical range of values from sorted set
        ``name`` between ``max`` and ``min``.

        If ``start`` and ``num`` are specified, then return a slice of the
        range.
        """

        if (start is not None and num is None) or (num is not None and start is None):
            raise RedisError("``start`` and ``num`` must both be specified")
        pieces = ["ZREVRANGEBYLEX", name, max, min]

        if start is not None and num is not None:
            pieces.extend([b("LIMIT"), start, num])

        return await self.execute_command(*pieces)

    async def zrangebyscore(
        self,
        name,
        min,
        max,
        start=None,
        num=None,
        withscores=False,
        score_cast_func=float,
    ):
        """
        Returns a range of values from the sorted set ``name`` with scores
        between ``min`` and ``max``.

        If ``start`` and ``num`` are specified, then return a slice
        of the range.

        ``withscores`` indicates to return the scores along with the values.
        The return type is a list of (value, score) pairs

        `score_cast_func`` a callable used to cast the score return value
        """

        if (start is not None and num is None) or (num is not None and start is None):
            raise RedisError("``start`` and ``num`` must both be specified")
        pieces = ["ZRANGEBYSCORE", name, min, max]

        if start is not None and num is not None:
            pieces.extend([b("LIMIT"), start, num])

        if withscores:
            pieces.append(b("WITHSCORES"))
        options = {"withscores": withscores, "score_cast_func": score_cast_func}

        return await self.execute_command(*pieces, **options)

    async def zrank(self, name, value):
        """
        Returns a 0-based value indicating the rank of ``value`` in sorted set
        ``name``
        """

        return await self.execute_command("ZRANK", name, value)

    async def zrem(self, name, *values):
        """Removes member ``values`` from sorted set ``name``"""

        return await self.execute_command("ZREM", name, *values)

    async def zremrangebylex(self, name, min, max):
        """
        Removes all elements in the sorted set ``name`` between the
        lexicographical range specified by ``min`` and ``max``.

        Returns the number of elements removed.
        """

        return await self.execute_command("ZREMRANGEBYLEX", name, min, max)

    async def zremrangebyrank(self, name, min, max):
        """
        Removes all elements in the sorted set ``name`` with ranks between
        ``min`` and ``max``. Values are 0-based, ordered from smallest score
        to largest. Values can be negative indicating the highest scores.
        Returns the number of elements removed
        """

        return await self.execute_command("ZREMRANGEBYRANK", name, min, max)

    async def zremrangebyscore(self, name, min, max):
        """
        Removes all elements in the sorted set ``name`` with scores
        between ``min`` and ``max``. Returns the number of elements removed.
        """

        return await self.execute_command("ZREMRANGEBYSCORE", name, min, max)

    async def zrevrange(
        self, name, start, end, withscores=False, score_cast_func=float
    ):
        """
        Returns a range of values from sorted set ``name`` between
        ``start`` and ``end`` sorted in descending order.

        ``start`` and ``end`` can be negative, indicating the end of the range.

        ``withscores`` indicates to return the scores along with the values
        The return type is a list of (value, score) pairs

        ``score_cast_func`` a callable used to cast the score return value
        """
        pieces = ["ZREVRANGE", name, start, end]

        if withscores:
            pieces.append(b("WITHSCORES"))
        options = {"withscores": withscores, "score_cast_func": score_cast_func}

        return await self.execute_command(*pieces, **options)

    async def zrangestore(
        self,
        dest,
        name,
        start,
        end,
        byscore=False,
        bylex=False,
        desc=False,
        offset=None,
        num=None,
    ):
        """
        Stores in ``dest`` the result of a range of values from sorted set
        ``name`` between ``start`` and ``end`` sorted in ascending order.
        ``start`` and ``end`` can be negative, indicating the end of the range.
        ``byscore`` when set to True, returns the range of elements from the
        sorted set having scores equal or between ``start`` and ``end``.
        ``bylex`` when set to True, returns the range of elements from the
        sorted set between the ``start`` and ``end`` lexicographical closed
        range intervals.
        Valid ``start`` and ``end`` must start with ( or [, in order to specify
        whether the range interval is exclusive or inclusive, respectively.
        ``desc`` a boolean indicating whether to sort the results in reversed
        order.
        ``offset`` and ``num`` are specified, then return a slice of the range.
        Can't be provided when using ``bylex``.

        .. versionadded:: 2.1.0
        """

        return await self._zrange(
            "ZRANGESTORE",
            dest,
            name,
            start,
            end,
            desc,
            byscore,
            bylex,
            False,
            None,
            offset,
            num,
        )

    async def zrevrangebyscore(
        self,
        name,
        max,
        min,
        start=None,
        num=None,
        withscores=False,
        score_cast_func=float,
    ):
        """
        Returns a range of values from the sorted set ``name`` with scores
        between ``min`` and ``max`` in descending order.

        If ``start`` and ``num`` are specified, then return a slice
        of the range.

        ``withscores`` indicates to return the scores along with the values.
        The return type is a list of (value, score) pairs

        ``score_cast_func`` a callable used to cast the score return value
        """

        if (start is not None and num is None) or (num is not None and start is None):
            raise RedisError("``start`` and ``num`` must both be specified")
        pieces = ["ZREVRANGEBYSCORE", name, max, min]

        if start is not None and num is not None:
            pieces.extend([b("LIMIT"), start, num])

        if withscores:
            pieces.append(b("WITHSCORES"))
        options = {"withscores": withscores, "score_cast_func": score_cast_func}

        return await self.execute_command(*pieces, **options)

    async def zrevrank(self, name, value):
        """
        Returns a 0-based value indicating the descending rank of
        ``value`` in sorted set ``name``
        """

        return await self.execute_command("ZREVRANK", name, value)

    async def zscore(self, name, value):
        "Return the score of element ``value`` in sorted set ``name``"

        return await self.execute_command("ZSCORE", name, value)

    async def zunion(self, keys, aggregate=None, withscores=False):
        """
        Return the union of multiple sorted sets specified by ``keys``.
        ``keys`` can be provided as dictionary of keys and their weights.
        Scores will be aggregated based on the ``aggregate``, or SUM if
        none is provided.

        .. versionadded:: 2.1.0
        """

        return await self._zaggregate(
            "ZUNION", None, keys, aggregate, withscores=withscores
        )

    async def zunionstore(self, dest, keys, aggregate=None):
        """
        Performs Union on multiple sorted sets specified by ``keys`` into
        a new sorted set, ``dest``. Scores in the destination will be
        aggregated based on the ``aggregate``, or SUM if none is provided.
        """

        return await self._zaggregate("ZUNIONSTORE", dest, keys, aggregate)

    async def zmscore(self, key, members):
        """
        Returns the scores associated with the specified members
        in the sorted set stored at key.
        ``members`` should be a list of the member name.
        Return type is a list of score.
        If the member does not exist, a None will be returned
        in corresponding position.

        .. versionadded:: 2.1.0
        """

        if not members:
            raise DataError("ZMSCORE members must be a non-empty list")
        pieces = [key] + members

        return await self.execute_command("ZMSCORE", *pieces)

    async def _zaggregate(self, command, dest, keys, aggregate=None):
        pieces = [command, dest, len(keys)]

        if isinstance(keys, dict):
            keys, weights = iterkeys(keys), itervalues(keys)
        else:
            weights = None
        pieces.extend(keys)

        if weights:
            pieces.append(b("WEIGHTS"))
            pieces.extend(weights)

        if aggregate:
            pieces.append(b("AGGREGATE"))
            pieces.append(aggregate)

        return await self.execute_command(*pieces)

    async def zscan(
        self, name, cursor=0, match=None, count=None, score_cast_func=float
    ):
        """
        Incrementally returns lists of elements in a sorted set. Also returns
        a cursor pointing to the scan position.

        ``match`` allows for filtering the keys by pattern

        ``count`` allows for hint the minimum number of returns

        ``score_cast_func`` a callable used to cast the score return value
        """
        pieces = [name, cursor]

        if match is not None:
            pieces.extend([b("MATCH"), match])

        if count is not None:
            pieces.extend([b("COUNT"), count])
        options = {"score_cast_func": score_cast_func}

        return await self.execute_command("ZSCAN", *pieces, **options)


class ClusterSortedSetCommandMixin(SortedSetCommandMixin):

    RESULT_CALLBACKS = {"ZSCAN": first_key}



from coredis.exceptions import DataError, RedisError
from coredis.utils import nativestr


def parse_geosearch_generic(response, **options):
    """
    Parse the response of 'GEOSEARCH', GEORADIUS' and 'GEORADIUSBYMEMBER'
    commands according to 'withdist', 'withhash' and 'withcoord' labels.
    """

    if options["store"] or options["store_dist"]:
        # `store` and `store_dist` cant be combined
        # with other command arguments.
        # relevant to 'GEORADIUS' and 'GEORADIUSBYMEMBER'

        return response

    if type(response) != list:
        response_list = [response]
    else:
        response_list = response

    if not options["withdist"] and not options["withcoord"] and not options["withhash"]:
        # just a bunch of places

        return [nativestr(r) for r in response_list]

    cast = {
        "withdist": float,
        "withcoord": lambda ll: (float(ll[0]), float(ll[1])),
        "withhash": int,
    }

    # zip all output results with each casting function to get
    # the properly native Python value.
    f = [nativestr]
    f += [cast[o] for o in ["withdist", "withhash", "withcoord"] if options[o]]

    return [list(map(lambda fv: fv[0](fv[1]), zip(f, r))) for r in response_list]


class GeoCommandMixin:

    RESPONSE_CALLBACKS = {
        "GEOPOS": lambda r: list(
            map(lambda ll: (float(ll[0]), float(ll[1])) if ll is not None else None, r)
        ),
        "GEOHASH": lambda r: list(r),
        "GEOSEARCH": parse_geosearch_generic,
        "GEORADIUS": parse_geosearch_generic,
        "GEORADIUSBYMEMBER": parse_geosearch_generic,
        "GEODIST": float,
        "GEOADD": int,
    }

    # GEO COMMANDS
    async def geoadd(self, name, *values):
        """
        Add the specified geospatial items to the specified key identified
        by the ``name`` argument. The Geospatial items are given as ordered
        members of the ``values`` argument, each item or place is formed by
        the triad latitude, longitude and name.
        """

        if len(values) % 3 != 0:
            raise RedisError("GEOADD requires places with lon, lat and name" " values")

        return await self.execute_command("GEOADD", name, *values)

    async def geodist(self, name, place1, place2, unit=None):
        """
        Return the distance between ``place1`` and ``place2`` members of the
        ``name`` key.
        The units must be one of the following : m, km mi, ft. By async default
        meters are used.
        """
        pieces = [name, place1, place2]

        if unit and unit not in ("m", "km", "mi", "ft"):
            raise RedisError("GEODIST invalid unit")
        elif unit:
            pieces.append(unit)

        return await self.execute_command("GEODIST", *pieces)

    async def geohash(self, name, *values):
        """
        Return the geo hash string for each item of ``values`` members of
        the specified key identified by the ``name`` argument.
        """

        return await self.execute_command("GEOHASH", name, *values)

    async def geopos(self, name, *values):
        """
        Return the positions of each item of ``values`` as members of
        the specified key identified by the ``name`` argument. Each position
        is represented by the pairs lon and lat.
        """

        return await self.execute_command("GEOPOS", name, *values)

    async def georadius(
        self,
        name,
        longitude,
        latitude,
        radius,
        unit=None,
        withdist=False,
        withcoord=False,
        withhash=False,
        count=None,
        sort=None,
        store=None,
        store_dist=None,
        any=False,
    ):
        """
        Return the members of the specified key identified by the
        ``name`` argument which are within the borders of the area specified
        with the ``latitude`` and ``longitude`` location and the maximum
        distance from the center specified by the ``radius`` value.

        The units must be one of the following : m, km mi, ft. By default

        ``withdist`` indicates to return the distances of each place.

        ``withcoord`` indicates to return the latitude and longitude of
        each place.

        ``withhash`` indicates to return the geohash string of each place.

        ``count`` indicates to return the number of elements up to N.

        ``sort`` indicates to return the places in a sorted way, ASC for
        nearest to fairest and DESC for fairest to nearest.

        ``store`` indicates to save the places names in a sorted set named
        with a specific key, each element of the destination sorted set is
        populated with the score got from the original geo sorted set.

        ``store_dist`` indicates to save the places names in a sorted set
        named with a specific key, instead of ``store`` the sorted set
        destination score is set with the distance.
        """

        return await self._georadiusgeneric(
            "GEORADIUS",
            name,
            longitude,
            latitude,
            radius,
            unit=unit,
            withdist=withdist,
            withcoord=withcoord,
            withhash=withhash,
            count=count,
            sort=sort,
            store=store,
            store_dist=store_dist,
            any=any,
        )

    async def georadiusbymember(
        self,
        name,
        member,
        radius,
        unit=None,
        withdist=False,
        withcoord=False,
        withhash=False,
        count=None,
        sort=None,
        store=None,
        store_dist=None,
        any=False,
    ):
        """
        This command is exactly like ``georadius`` with the sole difference
        that instead of taking, as the center of the area to query, a longitude
        and latitude value, it takes the name of a member already existing
        inside the geospatial index represented by the sorted set.
        """

        return await self._georadiusgeneric(
            "GEORADIUSBYMEMBER",
            name,
            member,
            radius,
            unit=unit,
            withdist=withdist,
            withcoord=withcoord,
            withhash=withhash,
            count=count,
            sort=sort,
            store=store,
            store_dist=store_dist,
            any=any,
        )

    async def _georadiusgeneric(self, command, *args, **kwargs):
        pieces = list(args)

        if kwargs["unit"] and kwargs["unit"] not in ("m", "km", "mi", "ft"):
            raise DataError("GEORADIUS invalid unit")
        elif kwargs["unit"]:
            pieces.append(kwargs["unit"])
        else:
            pieces.append(
                "m",
            )

        if kwargs["any"] and kwargs["count"] is None:
            raise DataError("``any`` can't be provided without ``count``")

        for arg_name, byte_repr in (
            ("withdist", "WITHDIST"),
            ("withcoord", "WITHCOORD"),
            ("withhash", "WITHHASH"),
        ):
            if kwargs[arg_name]:
                pieces.append(byte_repr)

        if kwargs["count"] is not None:
            pieces.extend(["COUNT", kwargs["count"]])

            if kwargs["any"]:
                pieces.append("ANY")

        if kwargs["sort"]:
            if kwargs["sort"] == "ASC":
                pieces.append("ASC")
            elif kwargs["sort"] == "DESC":
                pieces.append("DESC")
            else:
                raise DataError("GEORADIUS invalid sort")

        if kwargs["store"] and kwargs["store_dist"]:
            raise DataError("GEORADIUS store and store_dist cant be set" " together")

        if kwargs["store"]:
            pieces.extend([b"STORE", kwargs["store"]])

        if kwargs["store_dist"]:
            pieces.extend([b"STOREDIST", kwargs["store_dist"]])

        return await self.execute_command(command, *pieces, **kwargs)

    async def geosearch(
        self,
        name,
        member=None,
        longitude=None,
        latitude=None,
        unit="m",
        radius=None,
        width=None,
        height=None,
        sort=None,
        count=None,
        any=False,
        withcoord=False,
        withdist=False,
        withhash=False,
    ):
        """
        Return the members of specified key identified by the
        ``name`` argument, which are within the borders of the
        area specified by a given shape. This command extends the
        GEORADIUS command, so in addition to searching within circular
        areas, it supports searching within rectangular areas.
        This command should be used in place of the deprecated
        GEORADIUS and GEORADIUSBYMEMBER commands.

        ``member`` Use the position of the given existing
         member in the sorted set. Can't be given with ``longitude``
         and ``latitude``.

        ``longitude`` and ``latitude`` Use the position given by
         this coordinates. Can't be given with ``member``

        ``radius`` Similar to GEORADIUS, search inside circular
        area according the given radius. Can't be given with
        ``height`` and ``width``.

        ``height`` and ``width`` Search inside an axis-aligned
        rectangle, determined by the given height and width.
        Can't be given with ``radius``

        ``unit`` must be one of the following : m, km, mi, ft.
        `m` for meters (the default value), `km` for kilometers,
        `mi` for miles and `ft` for feet.

        ``sort`` indicates to return the places in a sorted way,
        ASC for nearest to farest and DESC for farest to nearest.

        ``count`` limit the results to the first count matching items.

        ``any`` is set to True, the command will return as soon as
        enough matches are found. Can't be provided without ``count``

        ``withdist`` indicates to return the distances of each place.

        ``withcoord`` indicates to return the latitude and longitude of
        each place.

        ``withhash`` indicates to return the geohash string of each place.

        .. versionadded:: 2.1.0
        """

        return await self._geosearchgeneric(
            "GEOSEARCH",
            name,
            member=member,
            longitude=longitude,
            latitude=latitude,
            unit=unit,
            radius=radius,
            width=width,
            height=height,
            sort=sort,
            count=count,
            any=any,
            withcoord=withcoord,
            withdist=withdist,
            withhash=withhash,
            store=None,
            store_dist=None,
        )

    async def geosearchstore(
        self,
        dest,
        name,
        member=None,
        longitude=None,
        latitude=None,
        unit="m",
        radius=None,
        width=None,
        height=None,
        sort=None,
        count=None,
        any=False,
        storedist=False,
    ):
        """
        This command is like GEOSEARCH, but stores the result in
        ``dest``. By default, it stores the results in the destination
        sorted set with their geospatial information.
        if ``store_dist`` set to True, the command will stores the
        items in a sorted set populated with their distance from the
        center of the circle or box, as a floating-point number.

        .. versionadded:: 2.1.0
        """

        return await self._geosearchgeneric(
            "GEOSEARCHSTORE",
            dest,
            name,
            member=member,
            longitude=longitude,
            latitude=latitude,
            unit=unit,
            radius=radius,
            width=width,
            height=height,
            sort=sort,
            count=count,
            any=any,
            withcoord=None,
            withdist=None,
            withhash=None,
            store=None,
            store_dist=storedist,
        )

    async def _geosearchgeneric(self, command, *args, **kwargs):
        pieces = list(args)

        # FROMMEMBER or FROMLONLAT

        if kwargs["member"] is None:
            if kwargs["longitude"] is None or kwargs["latitude"] is None:
                raise DataError(
                    "GEOSEARCH must have member or" " longitude and latitude"
                )

        if kwargs["member"]:
            if kwargs["longitude"] or kwargs["latitude"]:
                raise DataError(
                    "GEOSEARCH member and longitude or latitude" " cant be set together"
                )
            pieces.extend([b"FROMMEMBER", kwargs["member"]])

        if kwargs["longitude"] and kwargs["latitude"]:
            pieces.extend([b"FROMLONLAT", kwargs["longitude"], kwargs["latitude"]])

        # BYRADIUS or BYBOX

        if kwargs["radius"] is None:
            if kwargs["width"] is None or kwargs["height"] is None:
                raise DataError("GEOSEARCH must have radius or" " width and height")

        if kwargs["unit"] is None:
            raise DataError("GEOSEARCH must have unit")

        if kwargs["unit"].lower() not in ("m", "km", "mi", "ft"):
            raise DataError("GEOSEARCH invalid unit")

        if kwargs["radius"]:
            if kwargs["width"] or kwargs["height"]:
                raise DataError(
                    "GEOSEARCH radius and width or height" " cant be set together"
                )
            pieces.extend([b"BYRADIUS", kwargs["radius"], kwargs["unit"]])

        if kwargs["width"] and kwargs["height"]:
            pieces.extend([b"BYBOX", kwargs["width"], kwargs["height"], kwargs["unit"]])

        # sort

        if kwargs["sort"]:
            if kwargs["sort"].upper() == "ASC":
                pieces.append(b"ASC")
            elif kwargs["sort"].upper() == "DESC":
                pieces.append(b"DESC")
            else:
                raise DataError("GEOSEARCH invalid sort")

        # count any

        if kwargs["count"]:
            pieces.extend([b"COUNT", kwargs["count"]])

            if kwargs["any"]:
                pieces.append(b"ANY")
        elif kwargs["any"]:
            raise DataError("GEOSEARCH ``any`` can't be provided " "without count")

        # other properties

        for arg_name, byte_repr in (
            ("withdist", b"WITHDIST"),
            ("withcoord", b"WITHCOORD"),
            ("withhash", b"WITHHASH"),
            ("store_dist", b"STOREDIST"),
        ):
            if kwargs[arg_name]:
                pieces.append(byte_repr)

        return await self.execute_command(command, *pieces, **kwargs)



from coredis.exceptions import DataError
from coredis.utils import NodeFlag, bool_ok, dict_merge, list_keys_to_dict, nativestr


class ScriptingCommandMixin:

    RESPONSE_CALLBACKS = {
        "SCRIPT EXISTS": lambda r: list(map(bool, r)),
        "SCRIPT FLUSH": bool_ok,
        "SCRIPT KILL": bool_ok,
        "SCRIPT LOAD": nativestr,
    }

    async def eval(self, script, numkeys, *keys_and_args):
        """
        Execute the Lua ``script``, specifying the ``numkeys`` the script
        will touch and the key names and argument values in ``keys_and_args``.
        Returns the result of the script.

        In practice, use the object returned by ``register_script``. This
        function exists purely for Redis API completion.
        """

        return await self.execute_command("EVAL", script, numkeys, *keys_and_args)

    async def evalsha(self, sha, numkeys, *keys_and_args):
        """
        Use the ``sha`` to execute a Lua script already registered via EVAL
        or SCRIPT LOAD. Specify the ``numkeys`` the script will touch and the
        key names and argument values in ``keys_and_args``. Returns the result
        of the script.

        In practice, use the object returned by ``register_script``. This
        function exists purely for Redis API completion.
        """

        return await self.execute_command("EVALSHA", sha, numkeys, *keys_and_args)

    async def script_exists(self, *args):
        """
        Check if a script exists in the script cache by specifying the SHAs of
        each script as ``args``. Returns a list of boolean values indicating if
        if each already script exists in the cache.
        """

        return await self.execute_command("SCRIPT EXISTS", *args)

    async def script_flush(self, sync_type=None):
        """
        Flushes all scripts from the script cache

        :param sync_type: ``SYNC`` or ``ASYNC``. Default ``SYNC``

        .. versionadded:: 2.1.0
        """

        if sync_type not in ["SYNC", "ASYNC", None]:
            raise DataError(
                "SCRIPT FLUSH defaults to SYNC in redis > 6.2, or "
                "accepts SYNC/ASYNC. For older versions, "
                "of redis leave as None."
            )

        if sync_type is None:
            pieces = []
        else:
            pieces = [sync_type]

        return await self.execute_command("SCRIPT FLUSH", *pieces)

    async def script_kill(self):
        """Kills the currently executing Lua script"""

        return await self.execute_command("SCRIPT KILL")

    async def script_load(self, script):
        """Loads a Lua ``script`` into the script cache. Returns the SHA."""

        return await self.execute_command("SCRIPT LOAD", script)

    def register_script(self, script):
        """
        Registers a Lua ``script`` specifying the ``keys`` it will touch.
        Returns a Script object that is callable and hides the complexity of
        dealing with scripts, keys, and shas. This is the preferred way of
        working with Lua scripts.
        """
        from coredis.scripting import Script

        return Script(self, script)


class ClusterScriptingCommandMixin(ScriptingCommandMixin):

    NODES_FLAGS = dict_merge(
        {"SCRIPT KILL": NodeFlag.BLOCKED},
        list_keys_to_dict(
            ["SCRIPT LOAD", "SCRIPT FLUSH", "SCRIPT EXISTS"], NodeFlag.ALL_MASTERS
        ),
    )

    RESULT_CALLBACKS = {
        "SCRIPT LOAD": lambda res: list(res.values()).pop(),
        "SCRIPT EXISTS": lambda res: [all(k) for k in zip(*res.values())],
        "SCRIPT FLUSH": lambda res: all(res.values()),
    }



import inspect
import sys
from asyncio import CancelledError
from itertools import chain

from coredis.client import StrictRedis, StrictRedisCluster
from coredis.exceptions import (
    AskError,
    ClusterTransactionError,
    ConnectionError,
    ExecAbortError,
    MovedError,
    RedisClusterException,
    RedisError,
    ResponseError,
    TimeoutError,
    TryAgainError,
    WatchError,
)
from coredis.utils import clusterdown_wrapper, dict_merge

ERRORS_ALLOW_RETRY = (
    ConnectionError,
    TimeoutError,
    MovedError,
    AskError,
    TryAgainError,
)


class BasePipeline:
    """
    Pipelines provide a way to transmit multiple commands to the Redis server
    in one transmission.  This is convenient for batch processing, such as
    saving all the values in a list to Redis.

    All commands executed within a pipeline are wrapped with MULTI and EXEC
    calls. This guarantees all commands executed in the pipeline will be
    executed atomically.

    Any command raising an exception does *not* halt the execution of
    subsequent commands in the pipeline. Instead, the exception is caught
    and its instance is placed into the response list returned by execute().
    Code iterating over the response list should be able to deal with an
    instance of an exception as a potential value. In general, these will be
    ResponseError exceptions, such as those raised when issuing a command
    on a key of a different datatype.
    """

    UNWATCH_COMMANDS = {"DISCARD", "EXEC", "UNWATCH"}

    def __init__(self, connection_pool, response_callbacks, transaction, shard_hint):
        self.connection_pool = connection_pool
        self.connection = None
        self.response_callbacks = response_callbacks
        self.transaction = transaction
        self.shard_hint = shard_hint
        self.watching = False

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.reset()

    def __len__(self):
        return len(self.command_stack)

    async def reset(self):
        self.command_stack = []
        self.scripts = set()
        # make sure to reset the connection state in the event that we were
        # watching something
        if self.watching and self.connection:
            try:
                # call this manually since our unwatch or
                # immediate_execute_command methods can call reset()
                await self.connection.send_command("UNWATCH")
                await self.connection.read_response()
            except ConnectionError:
                # disconnect will also remove any previous WATCHes
                self.connection.disconnect()
        # clean up the other instance attributes
        self.watching = False
        self.explicit_transaction = False
        # we can safely return the connection to the pool here since we're
        # sure we're no longer WATCHing anything
        if self.connection:
            self.connection_pool.release(self.connection)
            self.connection = None

    def multi(self):
        """
        Starts a transactional block of the pipeline after WATCH commands
        are issued. End the transactional block with `execute`.
        """
        if self.explicit_transaction:
            raise RedisError("Cannot issue nested calls to MULTI")
        if self.command_stack:
            raise RedisError(
                "Commands without an initial WATCH have already " "been issued"
            )
        self.explicit_transaction = True

    async def execute_command(self, *args, **kwargs):
        if (self.watching or args[0] == "WATCH") and not self.explicit_transaction:
            return await self.immediate_execute_command(*args, **kwargs)
        return self.pipeline_execute_command(*args, **kwargs)

    async def immediate_execute_command(self, *args, **options):
        """
        Executes a command immediately, but don't auto-retry on a
        ConnectionError if we're already WATCHing a variable. Used when
        issuing WATCH or subsequent commands retrieving their values but before
        MULTI is called.
        """
        command_name = args[0]
        conn = self.connection
        # if this is the first call, we need a connection
        if not conn:
            conn = self.connection_pool.get_connection()
            self.connection = conn
        try:
            await conn.send_command(*args)
            return await self.parse_response(conn, command_name, **options)
        except (ConnectionError, TimeoutError) as e:
            conn.disconnect()
            if not conn.retry_on_timeout and isinstance(e, TimeoutError):
                raise
            # if we're not already watching, we can safely retry the command
            try:
                if not self.watching:
                    await conn.send_command(*args)
                    return await self.parse_response(conn, command_name, **options)
            except ConnectionError:
                # the retry failed so cleanup.
                conn.disconnect()
                await self.reset()
                raise

    def pipeline_execute_command(self, *args, **options):
        """
        Stages a command to be executed next execute() invocation

        Returns the current Pipeline object back so commands can be
        chained together, such as:

        pipe = pipe.set('foo', 'bar').incr('baz').decr('bang')

        At some other point, you can then run: pipe.execute(),
        which will execute all commands queued in the pipe.
        """
        self.command_stack.append((args, options))
        return self

    async def _execute_transaction(self, connection, commands, raise_on_error):
        cmds = chain([(("MULTI",), {})], commands, [(("EXEC",), {})])
        all_cmds = connection.pack_commands([args for args, _ in cmds])
        await connection.send_packed_command(all_cmds)
        errors = []

        # parse off the response for MULTI
        # NOTE: we need to handle ResponseErrors here and continue
        # so that we read all the additional command messages from
        # the socket
        try:
            await self.parse_response(connection, "_")
        except ResponseError:
            errors.append((0, sys.exc_info()[1]))

        # and all the other commands
        for i, command in enumerate(commands):
            try:
                await self.parse_response(connection, "_")
            except ResponseError:
                ex = sys.exc_info()[1]
                self.annotate_exception(ex, i + 1, command[0])
                errors.append((i, ex))

        # parse the EXEC.
        try:
            response = await self.parse_response(connection, "_")
        except ExecAbortError:
            if self.explicit_transaction:
                await self.immediate_execute_command("DISCARD")
            if errors:
                raise errors[0][1]
            raise sys.exc_info()[1]

        if response is None:
            raise WatchError("Watched variable changed.")

        # put any parse errors into the response
        for i, e in errors:
            response.insert(i, e)

        if len(response) != len(commands):
            self.connection.disconnect()
            raise ResponseError(
                "Wrong number of response items from " "pipeline execution"
            )

        # find any errors in the response and raise if necessary
        if raise_on_error:
            self.raise_first_error(commands, response)

        # We have to run response callbacks manually
        data = []
        for r, cmd in zip(response, commands):
            if not isinstance(r, Exception):
                args, options = cmd
                command_name = args[0]
                if command_name in self.response_callbacks:
                    callback = self.response_callbacks[command_name]
                    r = callback(r, **options)
                    # typing.Awaitable is not available in Python3.5
                    # so use inspect.isawaitable instead
                    # according to issue https://github.com/alisaifee/coredis/issues/77
                    if inspect.isawaitable(response):
                        r = await r
            data.append(r)
        return data

    async def _execute_pipeline(self, connection, commands, raise_on_error):
        # build up all commands into a single request to increase network perf
        all_cmds = connection.pack_commands([args for args, _ in commands])
        await connection.send_packed_command(all_cmds)

        response = []
        for args, options in commands:
            try:
                response.append(
                    await self.parse_response(connection, args[0], **options)
                )
            except ResponseError:
                response.append(sys.exc_info()[1])

        if raise_on_error:
            self.raise_first_error(commands, response)
        return response

    def raise_first_error(self, commands, response):
        for i, r in enumerate(response):
            if isinstance(r, ResponseError):
                self.annotate_exception(r, i + 1, commands[i][0])
                raise r

    def annotate_exception(self, exception, number, command):
        cmd = str(" ").join(map(str, command))
        msg = str("Command # %d (%s) of pipeline caused error: %s") % (
            number,
            cmd,
            str(exception.args[0]),
        )
        exception.args = (msg,) + exception.args[1:]

    async def _parse(self, connection, command_name, **options):
        "Parses a response from the Redis server"
        response = await connection.read_response()
        if command_name in self.response_callbacks:
            callback = self.response_callbacks[command_name]
            return callback(response, **options)
        return response

    async def parse_response(self, connection, command_name, **options):
        result = await self._parse(connection, command_name, **options)
        if command_name in self.UNWATCH_COMMANDS:
            self.watching = False
        elif command_name == "WATCH":
            self.watching = True
        return result

    async def load_scripts(self):
        # make sure all scripts that are about to be run on this pipeline exist
        scripts = list(self.scripts)
        immediate = self.immediate_execute_command
        shas = [s.sha for s in scripts]
        # we can't use the normal script_* methods because they would just
        # get buffered in the pipeline.
        exists = await immediate("SCRIPT EXISTS", *shas)
        if not all(exists):
            for s, exist in zip(scripts, exists):
                if not exist:
                    s.sha = await immediate("SCRIPT LOAD", s.script)

    async def execute(self, raise_on_error=True):
        """Executes all the commands in the current pipeline"""
        stack = self.command_stack
        if not stack:
            return []
        if self.scripts:
            await self.load_scripts()
        if self.transaction or self.explicit_transaction:
            exec = self._execute_transaction
        else:
            exec = self._execute_pipeline

        conn = self.connection
        if not conn:
            conn = self.connection_pool.get_connection()
            # assign to self.connection so reset() releases the connection
            # back to the pool after we're done
            self.connection = conn

        try:
            return await exec(conn, stack, raise_on_error)
        except (ConnectionError, TimeoutError, CancelledError) as e:
            conn.disconnect()
            if not conn.retry_on_timeout and isinstance(e, TimeoutError):
                raise
            # if we were watching a variable, the watch is no longer valid
            # since this connection has died. raise a WatchError, which
            # indicates the user should retry his transaction. If this is more
            # than a temporary failure, the WATCH that the user next issues
            # will fail, propegating the real ConnectionError
            if self.watching:
                raise WatchError(
                    "A ConnectionError occured on while watching " "one or more keys"
                )
            # otherwise, it's safe to retry since the transaction isn't
            # predicated on any state
            return await exec(conn, stack, raise_on_error)
        finally:
            await self.reset()

    async def watch(self, *names):
        """Watches the values at keys ``names``"""
        if self.explicit_transaction:
            raise RedisError("Cannot issue a WATCH after a MULTI")
        return await self.execute_command("WATCH", *names)

    async def unwatch(self):
        """Unwatches all previously specified keys"""
        return self.watching and await self.execute_command("UNWATCH") or True


class StrictPipeline(BasePipeline, StrictRedis):
    """Pipeline for the StrictRedis class"""


class StrictClusterPipeline(StrictRedisCluster):
    def __init__(
        self,
        connection_pool,
        result_callbacks=None,
        response_callbacks=None,
        startup_nodes=None,
        transaction=False,
        watches=None,
    ):
        self.command_stack = []
        self.refresh_table_asap = False
        self.connection_pool = connection_pool
        self.result_callbacks = (
            result_callbacks or self.__class__.RESULT_CALLBACKS.copy()
        )
        self.startup_nodes = startup_nodes if startup_nodes else []
        self.nodes_flags = self.__class__.NODES_FLAGS.copy()
        self.response_callbacks = dict_merge(
            response_callbacks or self.__class__.RESPONSE_CALLBACKS.copy()
        )
        self.transaction = transaction
        self.watches = watches or None
        self.watching = False
        self.explicit_transaction = False

    def __repr__(self):
        return "{0}".format(type(self).__name__)

    def __del__(self):
        self.reset()

    def __len__(self):
        return len(self.command_stack)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.reset()

    def _determine_slot(self, *args):
        """Figure out what slot based on command and args"""
        if len(args) <= 1:
            raise RedisClusterException(
                "No way to dispatch this command to Redis Cluster. Missing key."
            )
        command = args[0]

        if command in ["EVAL", "EVALSHA"]:
            numkeys = args[2]
            keys = args[3 : 3 + numkeys]
            slots = {self.connection_pool.nodes.keyslot(key) for key in keys}
            if len(slots) != 1:
                raise RedisClusterException(
                    "{0} - all keys must map to the same key slot".format(command)
                )
            return slots.pop()

        key = args[1]

        return self.connection_pool.nodes.keyslot(key)

    async def execute_command(self, *args, **kwargs):
        return self.pipeline_execute_command(*args, **kwargs)

    def pipeline_execute_command(self, *args, **options):
        self.command_stack.append(
            PipelineCommand(args, options, len(self.command_stack))
        )
        return self

    def raise_first_error(self, stack):
        for c in stack:
            r = c.result
            if isinstance(r, Exception):
                self.annotate_exception(r, c.position + 1, c.args)
                raise r

    def annotate_exception(self, exception, number, command):
        cmd = " ".join(str(x) for x in command)
        msg = "Command # {0} ({1}) of pipeline caused error: {2}".format(
            number, cmd, exception.args[0]
        )
        exception.args = (msg,) + exception.args[1:]

    async def execute(self, raise_on_error=True):
        await self.connection_pool.initialize()
        stack = self.command_stack

        if not stack:
            return []
        if self.transaction:
            execute = self.send_cluster_transaction
        else:
            execute = self.send_cluster_commands
        try:
            return await execute(stack, raise_on_error)
        finally:
            self.reset()

    def reset(self):
        """Empties pipeline"""
        self.command_stack = []

        self.scripts = set()
        self.watches = []
        # clean up the other instance attributes
        self.watching = False
        self.explicit_transaction = False

    @clusterdown_wrapper
    async def send_cluster_transaction(self, stack, raise_on_error=True):
        # the first time sending the commands we send all of the commands that were queued up.
        # if we have to run through it again, we only retry the commands that failed.
        attempt = sorted(stack, key=lambda x: x.position)
        node = {}

        # as we move through each command that still needs to be processed,
        # we figure out the slot number that command maps to, then from the slot determine the node.
        for c in attempt:
            # refer to our internal node -> slot table that tells us where a given
            # command should route to.
            slot = self._determine_slot(*c.args)
            hashed_node = self.connection_pool.get_node_by_slot(slot)

            # now that we know the name of the node ( it's just a string in the form of host:port )
            # we can build a list of commands for each node.
            if node.get("name") != hashed_node["name"]:
                # raise error if commands in a transaction can not hash to same node
                if len(node) > 0:
                    raise ClusterTransactionError(
                        "Keys in request don't hash to the same node"
                    )
            node = hashed_node
        conn = self.connection_pool.get_connection_by_node(node)
        if self.watches:
            await self._watch(node, conn, self.watches)
        node_commands = NodeCommands(self.parse_response, conn, in_transaction=True)
        node_commands.append(PipelineCommand(("MULTI",)))
        node_commands.extend(attempt)
        node_commands.append(PipelineCommand(("EXEC",)))
        self.explicit_transaction = True
        await node_commands.write()
        # todo: make this place clear
        try:
            await node_commands.read()
        except ExecAbortError:
            if self.explicit_transaction:
                await conn.send_command("DISCARD")
                await conn.read_response()

        # If at least one watched key is modified before the EXEC command,
        # the whole transaction aborts,
        # and EXEC returns a Null reply to notify that the transaction failed.
        if node_commands.commands[-1].result is None:
            raise WatchError
        self.connection_pool.release(conn)
        if self.watching:
            self._unwatch(conn)
        if raise_on_error:
            self.raise_first_error(stack)

    @clusterdown_wrapper
    async def send_cluster_commands(
        self, stack, raise_on_error=True, allow_redirections=True
    ):
        """
        Sends a bunch of cluster commands to the redis cluster.

        `allow_redirections` If the pipeline should follow `ASK` & `MOVED` responses
        automatically. If set to false it will raise RedisClusterException.
        """
        # the first time sending the commands we send all of the commands that were queued up.
        # if we have to run through it again, we only retry the commands that failed.
        attempt = sorted(stack, key=lambda x: x.position)

        # build a list of node objects based on node names we need to
        nodes = {}

        # as we move through each command that still needs to be processed,
        # we figure out the slot number that command maps to, then from the slot determine the node.
        for c in attempt:
            # refer to our internal node -> slot table that tells us where a given
            # command should route to.
            slot = self._determine_slot(*c.args)
            node = self.connection_pool.get_node_by_slot(slot)

            # little hack to make sure the node name is populated. probably could clean this up.
            self.connection_pool.nodes.set_node_name(node)

            # now that we know the name of the node ( it's just a string in the form of host:port )
            # we can build a list of commands for each node.
            node_name = node["name"]
            if node_name not in nodes:
                nodes[node_name] = NodeCommands(
                    self.parse_response,
                    self.connection_pool.get_connection_by_node(node),
                )

            nodes[node_name].append(c)

        # send the commands in sequence.
        # we  write to all the open sockets for each node first, before reading anything
        # this allows us to flush all the requests out across the network essentially in parallel
        # so that we can read them all in parallel as they come back.
        # we dont' multiplex on the sockets as they come available, but that shouldn't make
        # too much difference.
        node_commands = nodes.values()
        for n in node_commands:
            await n.write()

        for n in node_commands:
            await n.read()

        # release all of the redis connections we allocated earlier back into the connection pool.
        # we used to do this step as part of a try/finally block, but it is really dangerous to
        # release connections back into the pool if for some reason the socket has data still left
        # in it from a previous operation. The write and read operations already have try/catch
        # around them for all known types of errors including connection and socket level errors.
        # So if we hit an exception, something really bad happened and putting any of
        # these connections back into the pool is a very bad idea.
        # the socket might have unread buffer still sitting in it, and then the
        # next time we read from it we pass the buffered result back from a previous
        # command and every single request after to that connection will always get
        # a mismatched result. (not just theoretical, I saw this happen on production x.x).
        for n in nodes.values():
            self.connection_pool.release(n.connection)

        # if the response isn't an exception it is a valid response from the node
        # we're all done with that command, YAY!
        # if we have more commands to attempt, we've run into problems.
        # collect all the commands we are allowed to retry.
        # (MOVED, ASK, or connection errors or timeout errors)
        attempt = sorted(
            [c for c in attempt if isinstance(c.result, ERRORS_ALLOW_RETRY)],
            key=lambda x: x.position,
        )
        if attempt and allow_redirections:
            # RETRY MAGIC HAPPENS HERE!
            # send these remaing comamnds one at a time using `execute_command`
            # in the main client. This keeps our retry logic in one place mostly,
            # and allows us to be more confident in correctness of behavior.
            # at this point any speed gains from pipelining have been lost
            # anyway, so we might as well make the best attempt to get the correct
            # behavior.
            #
            # The client command will handle retries for each individual command
            # sequentially as we pass each one into `execute_command`. Any exceptions
            # that bubble out should only appear once all retries have been exhausted.
            #
            # If a lot of commands have failed, we'll be setting the
            # flag to rebuild the slots table from scratch. So MOVED errors should
            # correct themselves fairly quickly.
            await self.connection_pool.nodes.increment_reinitialize_counter(
                len(attempt)
            )
            for c in attempt:
                try:
                    # send each command individually like we do in the main client.
                    c.result = await super(StrictClusterPipeline, self).execute_command(
                        *c.args, **c.options
                    )
                except RedisError as e:
                    c.result = e

        # turn the response back into a simple flat array that corresponds
        # to the sequence of commands issued in the stack in pipeline.execute()
        response = [c.result for c in sorted(stack, key=lambda x: x.position)]

        if raise_on_error:
            self.raise_first_error(stack)

        return response

    def _fail_on_redirect(self, allow_redirections):
        if not allow_redirections:
            raise RedisClusterException(
                "ASK & MOVED redirection not allowed in this pipeline"
            )

    def _multi(self):
        raise RedisClusterException("method multi() is not implemented")

    def immediate_execute_command(self, *args, **options):
        raise RedisClusterException(
            "method immediate_execute_command() is not implemented"
        )

    def load_scripts(self):
        raise RedisClusterException("method load_scripts() is not implemented")

    async def _watch(self, node, conn, names):
        "Watches the values at keys ``names``"
        for name in names:
            slot = self._determine_slot("WATCH", name)
            dist_node = self.connection_pool.get_node_by_slot(slot)
            if node.get("name") != dist_node["name"]:
                # raise error if commands in a transaction can not hash to same node
                if len(node) > 0:
                    raise ClusterTransactionError(
                        "Keys in request don't hash to the same node"
                    )
        if self.explicit_transaction:
            raise RedisError("Cannot issue a WATCH after a MULTI")
        await conn.send_command("WATCH", *names)
        return await conn.read_response()

    async def _unwatch(self, conn):
        """Unwatches all previously specified keys"""
        await conn.send_command("UNWATCH")
        res = await conn.read_response()
        return self.watching and res or True

    def script_load_for_pipeline(self, *args, **kwargs):
        raise RedisClusterException(
            "method script_load_for_pipeline() is not implemented"
        )

    def delete(self, *names):
        """Deletes a key specified by ``names``"""
        if len(names) != 1:
            raise RedisClusterException(
                "deleting multiple keys is not implemented in pipeline command"
            )

        return self.execute_command("DEL", names[0])


def block_pipeline_command(func):
    """
    Prints error because some pipelined commands should be blocked when
    running in cluster-mode
    """

    def inner(*args, **kwargs):
        raise RedisClusterException(
            (
                f"ERROR: Calling pipelined function {func.__name__}"
                " is blocked when running redis in cluster mode..."
            )
        )

    return inner


# Blocked pipeline commands
StrictClusterPipeline.bgrewriteaof = block_pipeline_command(
    StrictClusterPipeline.bgrewriteaof
)
StrictClusterPipeline.bgsave = block_pipeline_command(StrictClusterPipeline.bgsave)
StrictClusterPipeline.bitop = block_pipeline_command(StrictClusterPipeline.bitop)
StrictClusterPipeline.brpoplpush = block_pipeline_command(
    StrictClusterPipeline.brpoplpush
)
StrictClusterPipeline.client_getname = block_pipeline_command(
    StrictClusterPipeline.client_getname
)
StrictClusterPipeline.client_kill = block_pipeline_command(
    StrictClusterPipeline.client_kill
)
StrictClusterPipeline.client_list = block_pipeline_command(
    StrictClusterPipeline.client_list
)
StrictClusterPipeline.client_setname = block_pipeline_command(
    StrictClusterPipeline.client_setname
)
StrictClusterPipeline.config_get = block_pipeline_command(
    StrictClusterPipeline.config_get
)
StrictClusterPipeline.config_resetstat = block_pipeline_command(
    StrictClusterPipeline.config_resetstat
)
StrictClusterPipeline.config_rewrite = block_pipeline_command(
    StrictClusterPipeline.config_rewrite
)
StrictClusterPipeline.config_set = block_pipeline_command(
    StrictClusterPipeline.config_set
)
StrictClusterPipeline.dbsize = block_pipeline_command(StrictClusterPipeline.dbsize)
StrictClusterPipeline.echo = block_pipeline_command(StrictClusterPipeline.echo)
StrictClusterPipeline.evalsha = block_pipeline_command(StrictClusterPipeline.evalsha)
StrictClusterPipeline.flushall = block_pipeline_command(StrictClusterPipeline.flushall)
StrictClusterPipeline.flushdb = block_pipeline_command(StrictClusterPipeline.flushdb)
StrictClusterPipeline.info = block_pipeline_command(StrictClusterPipeline.info)
StrictClusterPipeline.keys = block_pipeline_command(StrictClusterPipeline.keys)
StrictClusterPipeline.lastsave = block_pipeline_command(StrictClusterPipeline.lastsave)
StrictClusterPipeline.mget = block_pipeline_command(StrictClusterPipeline.mget)
StrictClusterPipeline.move = block_pipeline_command(StrictClusterPipeline.move)
StrictClusterPipeline.mset = block_pipeline_command(StrictClusterPipeline.mset)
StrictClusterPipeline.msetnx = block_pipeline_command(StrictClusterPipeline.msetnx)
StrictClusterPipeline.pfmerge = block_pipeline_command(StrictClusterPipeline.pfmerge)
StrictClusterPipeline.pfcount = block_pipeline_command(StrictClusterPipeline.pfcount)
StrictClusterPipeline.ping = block_pipeline_command(StrictClusterPipeline.ping)
StrictClusterPipeline.publish = block_pipeline_command(StrictClusterPipeline.publish)
StrictClusterPipeline.randomkey = block_pipeline_command(
    StrictClusterPipeline.randomkey
)
StrictClusterPipeline.rename = block_pipeline_command(StrictClusterPipeline.rename)
StrictClusterPipeline.renamenx = block_pipeline_command(StrictClusterPipeline.renamenx)
StrictClusterPipeline.rpoplpush = block_pipeline_command(
    StrictClusterPipeline.rpoplpush
)
StrictClusterPipeline.save = block_pipeline_command(StrictClusterPipeline.save)
StrictClusterPipeline.scan = block_pipeline_command(StrictClusterPipeline.scan)
StrictClusterPipeline.script_exists = block_pipeline_command(
    StrictClusterPipeline.script_exists
)
StrictClusterPipeline.script_flush = block_pipeline_command(
    StrictClusterPipeline.script_flush
)
StrictClusterPipeline.script_kill = block_pipeline_command(
    StrictClusterPipeline.script_kill
)
StrictClusterPipeline.script_load = block_pipeline_command(
    StrictClusterPipeline.script_load
)
StrictClusterPipeline.sdiff = block_pipeline_command(StrictClusterPipeline.sdiff)
StrictClusterPipeline.sdiffstore = block_pipeline_command(
    StrictClusterPipeline.sdiffstore
)
StrictClusterPipeline.sentinel_get_master_addr_by_name = block_pipeline_command(
    StrictClusterPipeline.sentinel_get_master_addr_by_name
)
StrictClusterPipeline.sentinel_master = block_pipeline_command(
    StrictClusterPipeline.sentinel_master
)
StrictClusterPipeline.sentinel_masters = block_pipeline_command(
    StrictClusterPipeline.sentinel_masters
)
StrictClusterPipeline.sentinel_monitor = block_pipeline_command(
    StrictClusterPipeline.sentinel_monitor
)
StrictClusterPipeline.sentinel_remove = block_pipeline_command(
    StrictClusterPipeline.sentinel_remove
)
StrictClusterPipeline.sentinel_sentinels = block_pipeline_command(
    StrictClusterPipeline.sentinel_sentinels
)
StrictClusterPipeline.sentinel_set = block_pipeline_command(
    StrictClusterPipeline.sentinel_set
)
StrictClusterPipeline.sentinel_slaves = block_pipeline_command(
    StrictClusterPipeline.sentinel_slaves
)
StrictClusterPipeline.shutdown = block_pipeline_command(StrictClusterPipeline.shutdown)
StrictClusterPipeline.sinter = block_pipeline_command(StrictClusterPipeline.sinter)
StrictClusterPipeline.sinterstore = block_pipeline_command(
    StrictClusterPipeline.sinterstore
)
StrictClusterPipeline.slaveof = block_pipeline_command(StrictClusterPipeline.slaveof)
StrictClusterPipeline.slowlog_get = block_pipeline_command(
    StrictClusterPipeline.slowlog_get
)
StrictClusterPipeline.slowlog_len = block_pipeline_command(
    StrictClusterPipeline.slowlog_len
)
StrictClusterPipeline.slowlog_reset = block_pipeline_command(
    StrictClusterPipeline.slowlog_reset
)
StrictClusterPipeline.smove = block_pipeline_command(StrictClusterPipeline.smove)
StrictClusterPipeline.sort = block_pipeline_command(StrictClusterPipeline.sort)
StrictClusterPipeline.sunion = block_pipeline_command(StrictClusterPipeline.sunion)
StrictClusterPipeline.sunionstore = block_pipeline_command(
    StrictClusterPipeline.sunionstore
)
StrictClusterPipeline.time = block_pipeline_command(StrictClusterPipeline.time)


class PipelineCommand:
    """
    TODO: document
    """

    def __init__(self, args, options=None, position=None):
        self.args = args
        if options is None:
            options = {}
        self.options = options
        self.position = position
        self.result = None
        self.node = None
        self.asking = False


class NodeCommands:
    """
    TODO: document
    """

    def __init__(self, parse_response, connection, in_transaction=False):
        self.parse_response = parse_response
        self.connection = connection
        self.commands = []
        self.in_transaction = in_transaction

    def extend(self, c):
        self.commands.extend(c)

    def append(self, c):
        self.commands.append(c)

    async def write(self):
        """NOTE: Code was borrowed from StrictRedis so it can be fixed"""
        connection = self.connection
        commands = self.commands

        # We are going to clobber the commands with the write, so go ahead
        # and ensure that nothing is sitting there from a previous run.
        for c in commands:
            c.result = None

        # build up all commands into a single request to increase network perf
        # send all the commands and catch connection and timeout errors.
        try:
            await connection.send_packed_command(
                connection.pack_commands([c.args for c in commands])
            )
        except (ConnectionError, TimeoutError) as e:
            for c in commands:
                c.result = e

    async def read(self):
        connection = self.connection
        for c in self.commands:

            # if there is a result on this command, it means we ran into an exception
            # like a connection error. Trying to parse a response on a connection that
            # is no longer open will result in a connection error raised by redis-py.
            # but redis-py doesn't check in parse_response that the sock object is
            # still set and if you try to read from a closed connection, it will
            # result in an AttributeError because it will do a readline() call on None.
            # This can have all kinds of nasty side-effects.
            # Treating this case as a connection error is fine because it will dump
            # the connection object back into the pool and on the next write, it will
            # explicitly open the connection and all will be well.
            if c.result is None:
                try:
                    if self.in_transaction:
                        cmd = "_"
                    else:
                        cmd = c.args[0]
                    c.result = await self.parse_response(connection, cmd, **c.options)
                except ExecAbortError:
                    raise
                except (ConnectionError, TimeoutError) as e:
                    for c in self.commands:
                        c.result = e
                except RedisError:
                    c.result = sys.exc_info()[1]



import datetime
import time as mod_time

from coredis.exceptions import DataError, RedisError, ResponseError
from coredis.utils import (
    NodeFlag,
    b,
    bool_ok,
    dict_merge,
    first_key,
    int_or_none,
    list_keys_to_dict,
    merge_result,
    string_keys_to_dict,
)


def sort_return_tuples(response, **options):
    """
    If ``groups`` is specified, return the response as a list of
    n-element tuples with n being the value found in options['groups']
    """

    if not response or not options["groups"]:
        return response
    n = options["groups"]

    return list(zip(*[response[i::n] for i in range(n)]))


def parse_object(response, infotype):
    """Parse the results of an OBJECT command"""

    if infotype in ("idletime", "refcount"):
        return int_or_none(response)

    return response


def parse_scan(response, **options):
    cursor, r = response

    return int(cursor), r


class KeysCommandMixin:

    RESPONSE_CALLBACKS = dict_merge(
        string_keys_to_dict("EXISTS EXPIRE EXPIREAT " "MOVE PERSIST RENAMENX", bool),
        {
            "DEL": int,
            "SORT": sort_return_tuples,
            "OBJECT": parse_object,
            "RANDOMKEY": lambda r: r and r or None,
            "SCAN": parse_scan,
            "RENAME": bool_ok,
        },
    )

    async def delete(self, *names):
        """Delete one or more keys specified by ``names``"""

        return await self.execute_command("DEL", *names)

    async def dump(self, name):
        """
        Return a serialized version of the value stored at the specified key.
        If key does not exist a nil bulk reply is returned.
        """

        return await self.execute_command("DUMP", name)

    async def object_encoding(self, name):
        """
        Return the internal encoding for the object stored at ``name``

        .. versionadded:: 2.1.0
        """

        return await self.execute_command("OBJECT ENCODING", name)

    async def object_freq(self, name):
        """
        Return the logarithmic access frequency counter for the object
        stored at ``name``

        .. versionadded:: 2.1.0
        """

        return await self.execute_command("OBJECT FREQ", name)

    async def object_idletime(self, name):
        """
        Return the time in seconds since the last access to the object
        stored at ``name``

        .. versionadded:: 2.1.0
        """

        return await self.execute_command("OBJECT IDLETIME", name)

    async def object_refcount(self, name):
        """
        Return the reference count of the object stored at ``name``

        .. versionadded:: 2.1.0
        """

        return await self.execute_command("OBJECT REFCOUNT", name)

    async def exists(self, name):
        """Returns a boolean indicating whether key ``name`` exists"""

        return await self.execute_command("EXISTS", name)

    async def expire(self, name, time):
        """
        Set an expire flag on key ``name`` for ``time`` seconds. ``time``
        can be represented by an integer or a Python timedelta object.
        """

        if isinstance(time, datetime.timedelta):
            time = time.seconds + time.days * 24 * 3600

        return await self.execute_command("EXPIRE", name, time)

    async def expireat(self, name, when):
        """
        Set an expire flag on key ``name``. ``when`` can be represented
        as an integer indicating unix time or a Python datetime object.
        """

        if isinstance(when, datetime.datetime):
            when = int(mod_time.mktime(when.timetuple()))

        return await self.execute_command("EXPIREAT", name, when)

    async def keys(self, pattern="*"):
        """Returns a list of keys matching ``pattern``"""

        return await self.execute_command("KEYS", pattern)

    async def move(self, name, db):
        """Moves the key ``name`` to a different Redis database ``db``"""

        return await self.execute_command("MOVE", name, db)

    async def object(self, infotype, key):
        """Returns the encoding, idletime, or refcount about the key"""

        return await self.execute_command("OBJECT", infotype, key, infotype=infotype)

    async def persist(self, name):
        """Removes an expiration on ``name``"""

        return await self.execute_command("PERSIST", name)

    async def pexpire(self, name, time):
        """
        Set an expire flag on key ``name`` for ``time`` milliseconds.
        ``time`` can be represented by an integer or a Python timedelta
        object.
        """

        if isinstance(time, datetime.timedelta):
            ms = int(time.microseconds / 1000)
            time = (time.seconds + time.days * 24 * 3600) * 1000 + ms

        return await self.execute_command("PEXPIRE", name, time)

    async def pexpireat(self, name, when):
        """
        Set an expire flag on key ``name``. ``when`` can be represented
        as an integer representing unix time in milliseconds (unix time * 1000)
        or a Python datetime object.
        """

        if isinstance(when, datetime.datetime):
            ms = int(when.microsecond / 1000)
            when = int(mod_time.mktime(when.timetuple())) * 1000 + ms

        return await self.execute_command("PEXPIREAT", name, when)

    async def pttl(self, name):
        """
        Returns the number of milliseconds until the key ``name`` will expire
        """

        return await self.execute_command("PTTL", name)

    async def randomkey(self):
        """Returns the name of a random key"""

        return await self.execute_command("RANDOMKEY")

    async def rename(self, src, dst):
        """
        Renames key ``src`` to ``dst``
        """

        return await self.execute_command("RENAME", src, dst)

    async def renamenx(self, src, dst):
        """Renames key ``src`` to ``dst`` if ``dst`` doesn't already exist"""

        return await self.execute_command("RENAMENX", src, dst)

    async def restore(self, name, ttl, value, replace=False):
        """
        Creates a key using the provided serialized value, previously obtained
        using DUMP.
        """
        params = [name, ttl, value]

        if replace:
            params.append("REPLACE")

        return await self.execute_command("RESTORE", *params)

    async def sort(
        self,
        name,
        start=None,
        num=None,
        by=None,
        get=None,
        desc=False,
        alpha=False,
        store=None,
        groups=False,
    ):
        """
        Sorts and returns a list, set or sorted set at ``name``.

        ``start`` and ``num`` are for paginating sorted data

        ``by`` allows using an external key to weight and sort the items.
            Use an "*" to indicate where in the key the item value is located

        ``get`` is for returning items from external keys rather than the
            sorted data itself.  Use an "*" to indicate where int he key
            the item value is located

        ``desc`` is for reversing the sort

        ``alpha`` is for sorting lexicographically rather than numerically

        ``store`` is for storing the result of the sort into
            the key ``store``

        ``groups`` if set to True and if ``get`` contains at least two
            elements, sort will return a list of tuples, each containing the
            values fetched from the arguments to ``get``.

        """

        if (start is not None and num is None) or (num is not None and start is None):
            raise RedisError("``start`` and ``num`` must both be specified")

        pieces = [name]

        if by is not None:
            pieces.append(b("BY"))
            pieces.append(by)

        if start is not None and num is not None:
            pieces.append(b("LIMIT"))
            pieces.append(start)
            pieces.append(num)

        if get is not None:
            # If get is a string assume we want to get a single value.
            # Otherwise assume it's an interable and we want to get multiple
            # values. We can't just iterate blindly because strings are
            # iterable.

            if isinstance(get, str):
                pieces.append(b("GET"))
                pieces.append(get)
            else:
                for g in get:
                    pieces.append(b("GET"))
                    pieces.append(g)

        if desc:
            pieces.append(b("DESC"))

        if alpha:
            pieces.append(b("ALPHA"))

        if store is not None:
            pieces.append(b("STORE"))
            pieces.append(store)

        if groups:
            if not get or isinstance(get, str) or len(get) < 2:
                raise DataError(
                    'when using "groups" the "get" argument '
                    "must be specified and contain at least "
                    "two keys"
                )

        options = {"groups": len(get) if groups else None}

        return await self.execute_command("SORT", *pieces, **options)

    async def touch(self, keys):
        """
        Alters the last access time of a key(s).
        A key is ignored if it does not exist.
        """

        return await self.execute_command("TOUCH", *keys)

    async def ttl(self, name):
        """Returns the number of seconds until the key ``name`` will expire"""

        return await self.execute_command("TTL", name)

    async def type(self, name):
        """Returns the type of key ``name``"""

        return await self.execute_command("TYPE", name)

    async def unlink(self, *keys):
        """Removes the specified keys in a different thread, not blocking"""

        return await self.execute_command("UNLINK", *keys)

    async def wait(self, num_replicas, timeout):
        """
        Redis synchronous replication
        That returns the number of replicas that processed the query when
        we finally have at least ``num_replicas``, or when the ``timeout`` was
        reached.
        """

        return await self.execute_command("WAIT", num_replicas, timeout)

    async def scan(self, cursor=0, match=None, count=None):
        """
        Incrementally return lists of key names. Also return a cursor
        indicating the scan position.

        ``match`` allows for filtering the keys by pattern

        ``count`` allows for hint the minimum number of returns
        """
        pieces = [cursor]

        if match is not None:
            pieces.extend([b("MATCH"), match])

        if count is not None:
            pieces.extend([b("COUNT"), count])

        return await self.execute_command("SCAN", *pieces)


class ClusterKeysCommandMixin(KeysCommandMixin):

    NODES_FLAGS = dict_merge(
        {
            "MOVE": NodeFlag.BLOCKED,
            "RANDOMKEY": NodeFlag.RANDOM,
            "SCAN": NodeFlag.ALL_MASTERS,
        },
        list_keys_to_dict(["KEYS"], NodeFlag.ALL_NODES),
    )

    RESULT_CALLBACKS = {
        "KEYS": merge_result,
        "RANDOMKEY": first_key,
        "SCAN": lambda res: res,
    }

    async def rename(self, src, dst):
        """
        Rename key ``src`` to ``dst``

        Cluster impl:
            This operation is no longer atomic because each key must be querried
            then set in separate calls because they maybe will change cluster node
        """

        if src == dst:
            raise ResponseError("source and destination objects are the same")

        data = await self.dump(src)

        if data is None:
            raise ResponseError("no such key")

        ttl = await self.pttl(src)

        if ttl is None or ttl < 1:
            ttl = 0

        await self.delete(dst)
        await self.restore(dst, ttl, data)
        await self.delete(src)

        return True

    async def delete(self, *names):
        """
        "Delete one or more keys specified by ``names``"

        Cluster impl:
            Iterate all keys and send DELETE for each key.
            This will go a lot slower than a normal delete call in StrictRedis.

            Operation is no longer atomic.
        """
        count = 0

        for arg in names:
            count += await self.execute_command("DEL", arg)

        return count

    async def renamenx(self, src, dst):
        """
        Rename key ``src`` to ``dst`` if ``dst`` doesn't already exist

        Cluster impl:
            Check if dst key do not exists, then calls rename().

            Operation is no longer atomic.
        """

        if not await self.exists(dst):
            return await self.rename(src, dst)

        return False
